"""
Object to manage regular expressions, try to optimize the result:
 - '(a|b)' => '[ab]'
 - '(color red|color blue)' => 'color (red|blue)'
 - '([ab]|c)' => '[abc]'
 - 'ab' + 'cd' => 'abcd' (one long string)
 - [a-z]|[b] => [a-z]
 - [a-c]|[a-e] => [a-z]
 - [a-c]|[d] => [a-d]
 - [a-c]|[d-f] => [a-f]

Operation:
 - str(): convert to string
 - repr(): debug string
 - a & b: concatenation, eg. "big " & "car" => "big car"
 - a + b: alias to a & b
 - a | b: a or b, eg. "dog" | "cat" => "dog|cat"
 - minLength(): minimum length of matching pattern, "(cat|horse)".minLength() => 3
 - maxLength(): maximum length of matching pattern, "(cat|horse)".maxLength() => 5

Utilities:
 - createString(): create a regex matching a string
 - createRange(): create a regex matching character ranges

TODO:
 - Support Unicode regex (avoid mixing str and unicode types)
 - createString("__tax") | parse("__[12]") => group '__'
 - Make sure that all RegexXXX() classes are inmutable
 - Use singleton for dot, start and end

See also CPAN Regexp::Assemble (Perl module):
   http://search.cpan.org/~dland/Regexp-Assemble-0.28/Assemble.pm
"""

import re
import operator

from hachoir.core.tools import makePrintable


def matchSingleValue(regex):
    """
    Regex only match one exact string.

    >>> matchSingleValue(RegexEmpty())
    True
    >>> matchSingleValue(createString("abc"))
    True
    >>> matchSingleValue(createRange("a", "b"))
    False
    >>> matchSingleValue(createRange("a"))
    True
    >>> matchSingleValue(RegexAnd((RegexStart(), createString("abc"))))
    True
    """
    cls = regex.__class__
    if cls in (RegexEmpty, RegexString, RegexStart, RegexEnd):
        return True
    if cls == RegexAnd:
        return all(matchSingleValue(item) for item in regex)
    if cls == RegexRange:
        return len(regex.ranges) == 1 and len(regex.ranges[0]) == 1
    return False


def escapeRegex(text):
    r"""
    Escape string to use it in a regular expression:
    prefix special characters « ^.+*?{}[]|()\$ » by an antislash.
    """
    return re.sub(r"([][^.+*?{}|()\\$])", r"\\\1", text)


def _join(func, regex_list):
    if not isinstance(regex_list, (tuple, list)):
        regex_list = list(regex_list)
    if len(regex_list) == 0:
        return RegexEmpty()
    regex = regex_list[0]
    for item in regex_list[1:]:
        regex = func(regex, item)
    return regex


def createString(text):
    """
    >>> createString('')
    <RegexEmpty ''>
    >>> createString('abc')
    <RegexString 'abc'>
    """
    if text:
        return RegexString(text)
    else:
        return RegexEmpty()


def createRange(*text, **kw):
    """
    Create a regex range using character list.

    >>> createRange("a", "d", "b")
    <RegexRange '[abd]'>
    >>> createRange("-", "9", "4", "3", "0")
    <RegexRange '[0349-]'>
    """
    ranges = (RegexRangeCharacter(item) for item in text)
    return RegexRange(ranges, kw.get('exclude', False))


class Regex:
    """
    Abstract class defining a regular expression atom
    """

    def minLength(self):
        """
        Maximum length in characters of the regex.
        Returns None if there is no limit.
        """
        raise NotImplementedError()

    def maxLength(self):
        """
        Maximum length in characters of the regex.
        Returns None if there is no limit.
        """
        return self.minLength()

    def __str__(self, **kw):
        if not hasattr(self, "_str_value"):
            self._str_value = {}
        key = kw.get('python', False)
        if key not in self._str_value:
            self._str_value[key] = self._str(**kw)
        return self._str_value[key]

    def _str(self, **kw):
        raise NotImplementedError()

    def __repr__(self, **kw):
        regex = self.__str__(**kw)
        regex = makePrintable(regex, 'ASCII')
        return "<%s '%s'>" % (
            self.__class__.__name__, regex)

    def __contains__(self, item):
        raise NotImplementedError()

    def match(self, other):
        """
        Guess if self may matchs regex.
        May returns False even if self does match regex.
        """
        if self == other:
            return True
        return self._match(other)

    def _match(self, other):
        """
        Does regex match other regex?
        Eg. "." matchs "0" or "[a-z]" but "0" doesn't match ".".

        This function is used by match() which already check regex identity.
        """
        return False

    def _and(self, regex):
        """
        Create new optimized version of a+b.
        Returns None if there is no interesting optimization.
        """
        return None

    def __and__(self, regex):
        """
        Create new optimized version of a & b.
        Returns None if there is no interesting optimization.

        >>> RegexEmpty() & RegexString('a')
        <RegexString 'a'>
        """
        if regex.__class__ == RegexEmpty:
            return self
        new_regex = self._and(regex)
        if new_regex:
            return new_regex
        else:
            return RegexAnd((self, regex))

    def __add__(self, regex):
        return self.__and__(regex)

    def or_(self, other):
        """
        Create new optimized version of a|b.
        Returns None if there is no interesting optimization.
        """

        # (a|a) => a
        if self == other:
            return self

        # a matchs b => a
        if self._match(other):
            return self

        # b matchs a => b
        if other._match(self):
            return other

        # Try to optimize (a|b)
        if self.__class__ != other.__class__:
            new_regex = self._or_(other, False)
            if new_regex:
                return new_regex

            # Try to optimize (b|a)
            new_regex = other._or_(self, True)
            if new_regex:
                return new_regex
            return None
        else:
            return self._or_(other, False)

    def _or_(self, other, reverse):
        """
        Try to create optimized version of self|other if reverse if False,
        or of other|self if reverse if True.
        """
        return None

    def __or__(self, other):
        """
        Public method of OR operator: a|b. It call or_() internal method.
        If or_() returns None: RegexOr object is used (and otherwise,
        use or_() result).
        """
        # Try to optimize (a|b)
        new_regex = self.or_(other)
        if new_regex:
            return new_regex

        # Else use (a|b)
        return RegexOr((self, other))

    def __eq__(self, regex):
        if self.__class__ != regex.__class__:
            return False
        return self._eq(regex)

    def _eq(self, other):
        """
        Check if two objects of the same class are equals
        """
        raise NotImplementedError(
            "Class %s has no method _eq()" % self.__class__.__name__)

    def compile(self, **kw):
        return re.compile(self.__str__(**kw))

    def findPrefix(self, regex):
        """
        Try to create a common prefix between two regex.
        Eg. "abc" and "abd" => "ab"

        Return None if no prefix can be found.
        """
        return None

    def __iter__(self):
        raise NotImplementedError()


class RegexEmpty(Regex):

    def minLength(self):
        return 0

    def _str(self, **kw):
        return ''

    def _and(self, other):
        return other

    def _eq(self, other):
        return True


class RegexWord(RegexEmpty):

    def _and(self, other):
        if other.__class__ == RegexWord:
            return self
        return None

    def _str(self, **kw):
        return r'\b'


class RegexStart(RegexEmpty):

    def _and(self, other):
        if other.__class__ == RegexStart:
            return self
        return None

    def _str(self, **kw):
        return '^'


class RegexEnd(RegexStart):

    def _and(self, other):
        if other.__class__ == RegexEnd:
            return self
        return None

    def _str(self, **kw):
        return '$'


class RegexDot(Regex):

    def minLength(self):
        return 1

    def _str(self, **kw):
        return '.'

    def _match(self, other):
        if other.__class__ == RegexRange:
            return True
        if other.__class__ == RegexString and len(other.text) == 1:
            return True
        return False

    def _eq(self, other):
        return True


class RegexString(Regex):

    def __init__(self, text=""):
        assert isinstance(text, str)
        self.text = text
        assert 1 <= len(self.text)

    def minLength(self):
        return len(self.text)

    def _and(self, regex):
        """
        >>> RegexString('a') + RegexString('b')
        <RegexString 'ab'>
        """
        if regex.__class__ == RegexString:
            return RegexString(self.text + regex.text)
        return None

    def _str(self, **kw):
        return escapeRegex(self.text)

    def findPrefix(self, regex):
        """
        Try to find a common prefix of two string regex, returns:
         - None if there is no common prefix
         - (prefix, regexa, regexb) otherwise => prefix + (regexa|regexb)

        >>> RegexString('color red').findPrefix(RegexString('color blue'))
        (<RegexString 'color '>, <RegexString 'red'>, <RegexString 'blue'>)
        """
        if regex.__class__ != RegexString:
            return None
        texta = self.text
        textb = regex.text

        # '(a|b)' => '[ab]'
        if len(texta) == len(textb) == 1:
            return (createRange(texta, textb), RegexEmpty(), RegexEmpty())

        # '(text abc|text def)' => 'text (abc|def)'
        common = None
        for length in range(1, min(len(texta), len(textb)) + 1):
            if textb.startswith(texta[:length]):
                common = length
            else:
                break
        if not common:
            return None
        return (RegexString(texta[:common]), createString(texta[common:]), createString(textb[common:]))

    def _or_(self, other, reverse):
        """
        Remove duplicate:
        >>> RegexString("color") | RegexString("color")
        <RegexString 'color'>

        Group prefix:

        >>> RegexString("color red") | RegexString("color blue")
        <RegexAnd 'color (red|blue)'>
        >>> RegexString("color red") | RegexString("color")
        <RegexAnd 'color( red|)'>

        """

        # Don't know any other optimization for str|other
        if other.__class__ != RegexString:
            return None

        # Find common prefix
        common = self.findPrefix(other)
        if common:
            if not reverse:
                regex = common[1] | common[2]
            else:
                regex = common[2] | common[1]
            return common[0] + regex
        return None

    def _eq(self, other):
        return self.text == other.text


class RegexRangeItem:

    def __init__(self, cmin, cmax=None):
        try:
            self.cmin = cmin
            if cmax is not None:
                self.cmax = cmax
            else:
                self.cmax = cmin
        except TypeError:
            raise TypeError("RegexRangeItem: two characters expected (%s, %s) found" % (
                type(cmin), type(cmax)))
        if self.cmax < self.cmin:
            raise TypeError("RegexRangeItem: minimum (%u) is bigger than maximum (%u)" %
                            (self.cmin, self.cmax))

    def __len__(self):
        return (self.cmax - self.cmin + 1)

    def __contains__(self, value):
        assert issubclass(value.__class__, RegexRangeItem)
        return (self.cmin <= value.cmin) and (value.cmax <= self.cmax)

    def __str__(self, **kw):
        cmin = chr(self.cmin)
        if self.cmin != self.cmax:
            cmax = chr(self.cmax)
            if (self.cmin + 1) == self.cmax:
                return "%s%s" % (cmin, cmax)
            else:
                return "%s-%s" % (cmin, cmax)
        else:
            return cmin

    def __repr__(self):
        return "<RegexRangeItem %u-%u>" % (self.cmin, self.cmax)


class RegexRangeCharacter(RegexRangeItem):

    def __init__(self, char):
        RegexRangeItem.__init__(self, ord(char), ord(char))


class RegexRange(Regex):

    def __init__(self, ranges, exclude=False, optimize=True):
        if optimize:
            self.ranges = []
            for item in ranges:
                RegexRange.rangeAdd(self.ranges, item)
            self.ranges.sort(key=lambda item: item.cmin)
        else:
            self.ranges = tuple(ranges)
        self.exclude = exclude

    @staticmethod
    def rangeAdd(ranges, itemb):
        """
        Add a value in a RegexRangeItem() list:
        remove duplicates and merge ranges when it's possible.
        """
        new = None
        for index, itema in enumerate(ranges):
            if itema in itemb:
                # [b] + [a-c] => [a-c]
                new = itemb
                break
            elif itemb in itema:
                # [a-c] + [b] => [a-c]
                return
            elif (itemb.cmax + 1) == itema.cmin:
                # [d-f] + [a-c] => [a-f]
                new = RegexRangeItem(itemb.cmin, itema.cmax)
                break
            elif (itema.cmax + 1) == itemb.cmin:
                # [a-c] + [d-f] => [a-f]
                new = RegexRangeItem(itema.cmin, itemb.cmax)
                break
        if new:
            del ranges[index]
            RegexRange.rangeAdd(ranges, new)
            return
        else:
            ranges.append(itemb)

    def minLength(self):
        return 1

    def _match(self, other):
        """
        >>> createRange("a") | createRange("b")
        <RegexRange '[ab]'>
        >>> createRange("a", "b", exclude=True) | createRange("a", "c", exclude=True)
        <RegexRange '[^a-c]'>
        """
        if not self.exclude and other.__class__ == RegexString and len(other.text) == 1:
            branges = (RegexRangeCharacter(other.text),)
        elif other.__class__ == RegexRange and self.exclude == other.exclude:
            branges = other.ranges
        else:
            return None
        for itemb in branges:
            if not any(itemb in itema for itema in self.ranges):
                return False
        return True

    def _or_(self, other, reverse):
        """
        >>> createRange("a") | createRange("b")
        <RegexRange '[ab]'>
        >>> createRange("a", "b", exclude=True) | createRange("a", "c", exclude=True)
        <RegexRange '[^a-c]'>
        """
        if not self.exclude and other.__class__ == RegexString and len(other.text) == 1:
            branges = (RegexRangeCharacter(other.text),)
        elif other.__class__ == RegexRange and self.exclude == other.exclude:
            branges = other.ranges
        else:
            return None
        ranges = list(self.ranges)
        for itemb in branges:
            RegexRange.rangeAdd(ranges, itemb)
        return RegexRange(ranges, self.exclude, optimize=False)

    def _str(self, **kw):
        content = [str(item) for item in self.ranges]
        if "-" in content:
            content.remove("-")
            suffix = "-"
        else:
            suffix = ""
        if "]" in content:
            content.remove("]")
            prefix = "]"
        else:
            prefix = ""
        text = prefix + (''.join(content)) + suffix
        if self.exclude:
            return "[^%s]" % text
        else:
            return "[%s]" % text

    def _eq(self, other):
        if self.exclude != other.exclude:
            return False
        return self.ranges == other.ranges


class RegexAnd(Regex):

    def __init__(self, items):
        self.content = list(items)
        assert 2 <= len(self.content)

    def _minmaxLength(self, lengths):
        total = 0
        for length in lengths:
            if length is None:
                return None
            total += length
        return total

    def minLength(self):
        """
        >>> regex=((RegexString('a') | RegexString('bcd')) + RegexString('z'))
        >>> regex.minLength()
        2
        """
        return self._minmaxLength(regex.minLength() for regex in self.content)

    def maxLength(self):
        """
        >>> regex=RegexOr((RegexString('a'), RegexString('bcd')))
        >>> RegexAnd((regex, RegexString('z'))).maxLength()
        4
        """
        return self._minmaxLength(regex.maxLength() for regex in self.content)

    def _or_(self, other, reverse):
        if other.__class__ == RegexString:
            contentb = [other]
        elif other.__class__ == RegexAnd:
            contentb = other.content
        else:
            return None

        contenta = self.content
        if reverse:
            contenta, contentb = contentb, contenta

        # Find common prefix
        # eg. (ab|ac) => a(b|c) and (abc|abd) => ab(c|d)
        index = 0
        last_index = min(len(contenta), len(contentb))
        while index < last_index and contenta[index] == contentb[index]:
            index += 1
        if index:
            regex = RegexAnd.join(
                contenta[index:]) | RegexAnd.join(contentb[index:])
            return RegexAnd.join(contenta[:index]) + regex

        # Find common prefix: (abc|aef) => a(bc|ef)
        common = contenta[0].findPrefix(contentb[0])
        if common:
            regexa = common[1] & RegexAnd.join(contenta[1:])
            regexb = common[2] & RegexAnd.join(contentb[1:])
            regex = (regexa | regexb)
            if matchSingleValue(common[0]) or matchSingleValue(regex):
                return common[0] + regex
        return None

    def _and(self, regex):
        """
        >>> RegexDot() + RegexDot()
        <RegexAnd '..'>
        >>> RegexDot() + RegexString('a') + RegexString('b')
        <RegexAnd '.ab'>
        """

        if regex.__class__ == RegexAnd:
            total = self
            for item in regex.content:
                total = total + item
            return total
        new_item = self.content[-1]._and(regex)
        if new_item:
            self.content[-1] = new_item
            return self
        return RegexAnd(self.content + [regex])

    def _str(self, **kw):
        return ''.join(item.__str__(**kw) for item in self.content)

    @classmethod
    def join(cls, regex):
        """
        >>> RegexAnd.join( (RegexString('Big '), RegexString('fish')) )
        <RegexString 'Big fish'>
        """
        return _join(operator.__and__, regex)

    def __iter__(self):
        return iter(self.content)

    def _eq(self, other):
        if len(self.content) != len(other.content):
            return False
        return all(item[0] == item[1] for item in zip(self.content, other.content))


class RegexOr(Regex):

    def __init__(self, items, optimize=True):
        if optimize:
            self.content = []
            for item in items:
                if item in self:
                    continue
                self.content.append(item)
        else:
            self.content = tuple(items)
        assert 2 <= len(self.content)

    def __contains__(self, regex):
        for item in self.content:
            if item == regex:
                return True
        return False

    def _or_(self, other, reverse):
        """
        >>> (RegexString("abc") | RegexString("123")) | (RegexString("plop") | RegexString("456"))
        <RegexOr '(abc|123|plop|456)'>
        >>> RegexString("mouse") | createRange('a') | RegexString("2006") | createRange('z')
        <RegexOr '(mouse|[az]|2006)'>
        """
        if other.__class__ == RegexOr:
            total = self
            for item in other.content:
                total = total | item
            return total
        for index, item in enumerate(self.content):
            new_item = item.or_(other)
            if new_item:
                content = list(self.content)
                content = content[:index] + [new_item] + content[index + 1:]
                return RegexOr(content, optimize=False)
        if not reverse:
            content = list(self.content) + [other]
        else:
            content = [other] + list(self.content)
        return RegexOr(content, optimize=False)

    def _str(self, **kw):
        content = '|'.join(item.__str__(**kw) for item in self.content)
        if kw.get('python', False):
            return "(?:%s)" % content
        else:
            return "(%s)" % content

    def _minmaxLength(self, lengths, func):
        value = None
        for length in lengths:
            if length is None:
                return None
            if value is None:
                value = length
            else:
                value = func(value, length)
        return value

    def minLength(self):
        lengths = (regex.minLength() for regex in self.content)
        return self._minmaxLength(lengths, min)

    def maxLength(self):
        lengths = (regex.maxLength() for regex in self.content)
        return self._minmaxLength(lengths, max)

    @classmethod
    def join(cls, regex):
        """
        >>> RegexOr.join( (RegexString('a'), RegexString('b'), RegexString('c')) )
        <RegexRange '[a-c]'>
        """
        return _join(operator.__or__, regex)

    def __iter__(self):
        return iter(self.content)

    def _eq(self, other):
        if len(self.content) != len(other.content):
            return False
        return all(item[0] == item[1] for item in zip(self.content, other.content))


def optimizeRepeatOr(rmin, rmax, regex):
    # Fix rmin/rmax
    for item in regex:
        cls = item.__class__
        if cls == RegexEmpty:
            # (a|b|){x,y} => (a|b){0,y}
            rmin = 0
        elif cls == RegexRepeat:
            # (a{0,n}|b){x,y} => (a{1,n}|b){0,y}
            if item.min == 0 and rmin == 1:
                rmin = 0

    # Create new (optimized) RegexOr expression
    content = []
    for item in regex:
        cls = item.__class__
        if cls == RegexEmpty:
            # (a|){x,y} => a{0,y}
            continue
        if cls == RegexRepeat:
            if item.min == 0:
                if rmin in (0, 1):
                    if rmax is item.max is None:
                        # (a*|b){x,} => (a|b){x,}
                        item = item.regex
                    else:
                        # (a{0,p}|b){x,} => (a{1,p}|b){x,}
                        item = RegexRepeat(
                            item.regex, 1, item.max, optimize=False)
            elif item.min == 1:
                if rmax is item.max is None:
                    # (a+|b){x,} => (a|b){x,}
                    item = item.regex
            else:
                if rmax is item.max is None:
                    # (a{n,}|b){x,} => (a{n}|b){x,}
                    item = RegexRepeat(item.regex, item.min,
                                       item.min, optimize=False)
        content.append(item)
    regex = RegexOr.join(content)
    return (rmin, rmax, regex)


class RegexRepeat(Regex):
    """
    >>> a=createString('a')
    >>> RegexRepeat(a, 0, None)
    <RegexRepeat 'a*'>
    >>> RegexRepeat(a, 1, None)
    <RegexRepeat 'a+'>
    >>> RegexRepeat(a, 0, 1)
    <RegexRepeat 'a?'>
    >>> RegexRepeat(a, 0, 1)
    <RegexRepeat 'a?'>
    >>> RegexRepeat(a, 1, 3)
    <RegexRepeat 'a{1,3}'>
    """

    def __init__(self, regex, rmin, rmax, optimize=True):
        # Optimisations
        if optimize:
            cls = regex.__class__
            if cls == RegexRepeat:
                # (a{n,p}){x,y) => a{n*x,p*y}
                if not (rmin == 0 and rmax == 1):
                    rmin *= regex.min
                    if regex.max and rmax:
                        rmax *= regex.max
                    else:
                        rmax = None
                    regex = regex.regex
            elif cls == RegexOr:
                rmin, rmax, regex = optimizeRepeatOr(rmin, rmax, regex)

        # Store attributes
        self.regex = regex
        self.min = rmin
        self.max = rmax

        # Post-conditions
        assert 0 <= rmin
        if self.max is not None:
            if self.max < self.min:
                raise ValueError(
                    "RegexRepeat: minimum (%s) is bigger than maximum (%s)!" % (self.min, self.max))
            if (self.max == 0) \
                    or (self.min == self.max == 1):
                raise ValueError(
                    "RegexRepeat: invalid values (min=%s, max=%s)!" % (self.min, self.max))

    def minLength(self):
        """
        >>> r=RegexRepeat(createString("abc") | createString("01"), 1, 3)
        >>> r.minLength(), r.maxLength()
        (2, 9)
        >>> r=RegexRepeat(createString("abc") | createString("01"), 4, None)
        >>> r.minLength(), r.maxLength()
        (8, None)
        """
        if self.min is not None:
            return self.regex.minLength() * self.min
        else:
            return None

    def maxLength(self):
        if self.max is not None:
            return self.regex.maxLength() * self.max
        else:
            return None

    def _str(self, **kw):
        text = str(self.regex)
        if self.regex.__class__ == RegexAnd \
                or (self.regex.__class__ == RegexString and 1 < len(self.regex.text)):
            text = "(%s)" % text
        if self.min == 0 and self.max == 1:
            return "%s?" % text
        if self.min == self.max:
            return "%s{%u}" % (text, self.min)
        if self.max is None:
            if self.min == 0:
                return "%s*" % text
            elif self.min == 1:
                return "%s+" % text
            else:
                return "%s{%u,}" % (text, self.min)
        return "%s{%u,%u}" % (text, self.min, self.max)

    def _eq(self, other):
        if self.min != other.min:
            return False
        if self.max != other.max:
            return False
        return (self.regex == other.regex)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Microsoft Windows Help (HLP) parser for Hachoir project.

Documents:
- Windows Help File Format / Annotation File Format / SHG and MRB File Format
  written by M. Winterhoff (100326.2776@compuserve.com)
  found on http://www.wotsit.org/

Author: Victor Stinner
Creation date: 2007-09-03
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           Bits, Int32, UInt16, UInt32,
                           NullBytes, RawBytes, PaddingBytes, String)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.text_handler import (textHandler, hexadecimal,
                                       displayHandler, humanFilesize)


class FileEntry(FieldSet):

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = self["res_space"].value * 8

    def createFields(self):
        yield displayHandler(UInt32(self, "res_space", "Reserved space"), humanFilesize)
        yield displayHandler(UInt32(self, "used_space", "Used space"), humanFilesize)
        yield Bits(self, "file_flags", 8, "(=4)")

        yield textHandler(UInt16(self, "magic"), hexadecimal)
        yield Bits(self, "flags", 16)
        yield displayHandler(UInt16(self, "page_size", "Page size in bytes"), humanFilesize)
        yield String(self, "structure", 16, strip="\0", charset="ASCII")
        yield NullBytes(self, "zero", 2)
        yield UInt16(self, "nb_page_splits", "Number of page splits B+ tree has suffered")
        yield UInt16(self, "root_page", "Page number of B+ tree root page")
        yield PaddingBytes(self, "one", 2, pattern="\xFF")
        yield UInt16(self, "nb_page", "Number of B+ tree pages")
        yield UInt16(self, "nb_level", "Number of levels of B+ tree")
        yield UInt16(self, "nb_entry", "Number of entries in B+ tree")

        size = (self.size - self.current_size) // 8
        if size:
            yield PaddingBytes(self, "reserved_space", size)


class HlpFile(Parser):
    PARSER_TAGS = {
        "id": "hlp",
        "category": "misc",
        "file_ext": ("hlp",),
        "min_size": 32,
        "description": "Microsoft Windows Help (HLP)",
    }
    endian = LITTLE_ENDIAN

    def validate(self):
        if self["magic"].value != 0x00035F3F:
            return "Invalid magic"
        if self["filesize"].value != self.stream.size // 8:
            return "Invalid magic"
        return True

    def createFields(self):
        yield textHandler(UInt32(self, "magic"), hexadecimal)
        yield UInt32(self, "dir_start", "Directory start")
        yield Int32(self, "first_free_block", "First free block")
        yield UInt32(self, "filesize", "File size in bytes")

        yield self.seekByte(self["dir_start"].value)
        yield FileEntry(self, "file[]")

        size = (self.size - self.current_size) // 8
        if size:
            yield RawBytes(self, "end", size)



"""Detailed ZLIB parser

Author: Robert Xiao
Creation date: July 9 2007

"""

from hachoir.parser import Parser
from hachoir.field import (Bit, Bits, Field, Int16, UInt32,
                           Enum, FieldSet, GenericFieldSet,
                           PaddingBits, ParserError, RawBytes)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal
from hachoir.core.tools import paddingSize, alignValue


def extend_data(data: bytearray, length, offset):
    """Extend data using a length and an offset, LZ-style."""
    if length >= offset:
        new_data = data[-offset:] * (alignValue(length, offset) // offset)
        data += new_data[:length]
    else:
        data += data[-offset:-offset + length]


def build_tree(lengths):
    """Build a Huffman tree from a list of lengths.
       The ith entry of the input list is the length of the Huffman code corresponding to
       integer i, or 0 if the integer i is unused."""
    max_length = max(lengths) + 1
    bit_counts = [0] * max_length
    next_code = [0] * max_length
    tree = {}
    for i in lengths:
        if i:
            bit_counts[i] += 1
    code = 0
    for i in range(1, len(bit_counts)):
        next_code[i] = code = (code + bit_counts[i - 1]) << 1
    for i, ln in enumerate(lengths):
        if ln:
            tree[(ln, next_code[ln])] = i
            next_code[ln] += 1
    return tree


class HuffmanCode(Field):
    """Huffman code. Uses tree parameter as the Huffman tree."""

    def __init__(self, parent, name, tree, description=None):
        Field.__init__(self, parent, name, 0, description)

        endian = self.parent.endian
        stream = self.parent.stream
        addr = self.absolute_address

        value = 0
        while (self.size, value) not in tree:
            if self.size > 256:
                raise ParserError("Huffman code too long!")
            bit = stream.readBits(addr, 1, endian)
            value <<= 1
            value += bit
            self._size += 1
            addr += 1
        self.huffvalue = value
        self.realvalue = tree[(self.size, value)]

    def createValue(self):
        return self.huffvalue


class DeflateBlock(FieldSet):
    # code: (min, max, extrabits)
    LENGTH_SYMBOLS = {257: (3, 3, 0),
                      258: (4, 4, 0),
                      259: (5, 5, 0),
                      260: (6, 6, 0),
                      261: (7, 7, 0),
                      262: (8, 8, 0),
                      263: (9, 9, 0),
                      264: (10, 10, 0),
                      265: (11, 12, 1),
                      266: (13, 14, 1),
                      267: (15, 16, 1),
                      268: (17, 18, 1),
                      269: (19, 22, 2),
                      270: (23, 26, 2),
                      271: (27, 30, 2),
                      272: (31, 34, 2),
                      273: (35, 42, 3),
                      274: (43, 50, 3),
                      275: (51, 58, 3),
                      276: (59, 66, 3),
                      277: (67, 82, 4),
                      278: (83, 98, 4),
                      279: (99, 114, 4),
                      280: (115, 130, 4),
                      281: (131, 162, 5),
                      282: (163, 194, 5),
                      283: (195, 226, 5),
                      284: (227, 257, 5),
                      285: (258, 258, 0)
                      }
    DISTANCE_SYMBOLS = {0: (1, 1, 0),
                        1: (2, 2, 0),
                        2: (3, 3, 0),
                        3: (4, 4, 0),
                        4: (5, 6, 1),
                        5: (7, 8, 1),
                        6: (9, 12, 2),
                        7: (13, 16, 2),
                        8: (17, 24, 3),
                        9: (25, 32, 3),
                        10: (33, 48, 4),
                        11: (49, 64, 4),
                        12: (65, 96, 5),
                        13: (97, 128, 5),
                        14: (129, 192, 6),
                        15: (193, 256, 6),
                        16: (257, 384, 7),
                        17: (385, 512, 7),
                        18: (513, 768, 8),
                        19: (769, 1024, 8),
                        20: (1025, 1536, 9),
                        21: (1537, 2048, 9),
                        22: (2049, 3072, 10),
                        23: (3073, 4096, 10),
                        24: (4097, 6144, 11),
                        25: (6145, 8192, 11),
                        26: (8193, 12288, 12),
                        27: (12289, 16384, 12),
                        28: (16385, 24576, 13),
                        29: (24577, 32768, 13),
                        }
    CODE_LENGTH_ORDER = [16, 17, 18, 0, 8, 7, 9,
                         6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15]

    def __init__(self, parent, name, uncomp_data=b"", *args, **kwargs):
        FieldSet.__init__(self, parent, name, *args, **kwargs)
        self.uncomp_data = bytearray(uncomp_data)

    def createFields(self):
        yield Bit(self, "final", "Is this the final block?")  # BFINAL
        yield Enum(Bits(self, "compression_type", 2),  # BTYPE
                   {0: "None", 1: "Fixed Huffman", 2: "Dynamic Huffman", 3: "Reserved"})
        if self["compression_type"].value == 0:  # no compression
            # align on byte boundary
            padding = paddingSize(self.current_size + self.absolute_address, 8)
            if padding:
                yield PaddingBits(self, "padding[]", padding)
            yield Int16(self, "len")
            yield Int16(self, "nlen", "One's complement of len")
            if self["len"].value != ~self["nlen"].value:
                raise ParserError(
                    "len must be equal to the one's complement of nlen!")
            # null stored blocks produced by some encoders (e.g. PIL)
            if self["len"].value:
                yield RawBytes(self, "data", self["len"].value, "Uncompressed data")
            return
        elif self["compression_type"].value == 1:  # Fixed Huffman
            length_tree = {}  # (size, huffman code): value
            distance_tree = {}
            for i in range(144):
                length_tree[(8, i + 48)] = i
            for i in range(144, 256):
                length_tree[(9, i + 256)] = i
            for i in range(256, 280):
                length_tree[(7, i - 256)] = i
            for i in range(280, 288):
                length_tree[(8, i - 88)] = i
            for i in range(32):
                distance_tree[(5, i)] = i
        elif self["compression_type"].value == 2:  # Dynamic Huffman
            yield Bits(self, "huff_num_length_codes", 5, "Number of Literal/Length Codes, minus 257")
            yield Bits(self, "huff_num_distance_codes", 5, "Number of Distance Codes, minus 1")
            yield Bits(self, "huff_num_code_length_codes", 4, "Number of Code Length Codes, minus 4")
            code_length_code_lengths = [0] * 19  # confusing variable name...
            for i in self.CODE_LENGTH_ORDER[:self["huff_num_code_length_codes"].value + 4]:
                field = Bits(
                    self, "huff_code_length_code[%i]" % i, 3, "Code lengths for the code length alphabet")
                yield field
                code_length_code_lengths[i] = field.value
            code_length_tree = build_tree(code_length_code_lengths)
            length_code_lengths = []
            distance_code_lengths = []
            for numcodes, name, lengths in (
                (self["huff_num_length_codes"].value +
                 257, "length", length_code_lengths),
                    (self["huff_num_distance_codes"].value + 1, "distance", distance_code_lengths)):
                while len(lengths) < numcodes:
                    field = HuffmanCode(self, "huff_%s_code[]" %
                                        name, code_length_tree)
                    value = field.realvalue
                    if value < 16:
                        prev_value = value
                        field._description = "Literal Code Length %i (Huffman Code %i)" % (
                            value, field.value)
                        yield field
                        lengths.append(value)
                    else:
                        info = {16: (3, 6, 2),
                                17: (3, 10, 3),
                                18: (11, 138, 7)}[value]
                        if value == 16:
                            repvalue = prev_value
                        else:
                            repvalue = 0
                        field._description = "Repeat Code %i, Repeating value (%i) %i to %i times (Huffman Code %i)" % (
                            value, repvalue, info[0], info[1], field.value)
                        yield field
                        extrafield = Bits(self, "huff_%s_code_extra[%s" % (
                            name, field.name.split('[')[1]), info[2])
                        num_repeats = extrafield.value + info[0]
                        extrafield._description = "Repeat Extra Bits (%i), total repeats %i" % (
                            extrafield.value, num_repeats)
                        yield extrafield
                        lengths += [repvalue] * num_repeats
            length_tree = build_tree(length_code_lengths)
            distance_tree = build_tree(distance_code_lengths)
        else:
            raise ParserError("Unsupported compression type 3!")
        while True:
            field = HuffmanCode(self, "length_code[]", length_tree)
            value = field.realvalue
            if value < 256:
                field._description = "Literal Code %r (Huffman Code %i)" % (
                    chr(value), field.value)
                yield field
                self.uncomp_data.append(value)
            if value == 256:
                field._description = "Block Terminator Code (256) (Huffman Code %i)" % field.value
                yield field
                break
            elif value > 256:
                info = self.LENGTH_SYMBOLS[value]
                if info[2] == 0:
                    field._description = "Length Code %i, Value %i (Huffman Code %i)" % (
                        value, info[0], field.value)
                    length = info[0]
                    yield field
                else:
                    field._description = "Length Code %i, Values %i to %i (Huffman Code %i)" % (
                        value, info[0], info[1], field.value)
                    yield field
                    extrafield = Bits(
                        self, "length_extra[%s" % field.name.split('[')[1], info[2])
                    length = extrafield.value + info[0]
                    extrafield._description = "Length Extra Bits (%i), total length %i" % (
                        extrafield.value, length)
                    yield extrafield
                field = HuffmanCode(self, "distance_code[]", distance_tree)
                value = field.realvalue
                info = self.DISTANCE_SYMBOLS[value]
                if info[2] == 0:
                    field._description = "Distance Code %i, Value %i (Huffman Code %i)" % (
                        value, info[0], field.value)
                    distance = info[0]
                    yield field
                else:
                    field._description = "Distance Code %i, Values %i to %i (Huffman Code %i)" % (
                        value, info[0], info[1], field.value)
                    yield field
                    extrafield = Bits(
                        self, "distance_extra[%s" % field.name.split('[')[1], info[2])
                    distance = extrafield.value + info[0]
                    extrafield._description = "Distance Extra Bits (%i), total length %i" % (
                        extrafield.value, distance)
                    yield extrafield
                extend_data(self.uncomp_data, length, distance)


class DeflateData(GenericFieldSet):
    endian = LITTLE_ENDIAN

    def createFields(self):
        uncomp_data = bytearray()
        blk = DeflateBlock(self, "compressed_block[]", uncomp_data)
        yield blk
        uncomp_data = blk.uncomp_data
        while not blk["final"].value:
            blk = DeflateBlock(self, "compressed_block[]", uncomp_data)
            yield blk
            uncomp_data = blk.uncomp_data
        # align on byte boundary
        padding = paddingSize(self.current_size + self.absolute_address, 8)
        if padding:
            yield PaddingBits(self, "padding[]", padding)
        self.uncompressed_data = uncomp_data


class ZlibData(Parser):
    PARSER_TAGS = {
        "id": "zlib",
        "category": "archive",
        "file_ext": ("zlib",),
        "min_size": 8 * 8,
        "description": "ZLIB Data",
    }
    endian = LITTLE_ENDIAN

    def validate(self):
        if self["compression_method"].value != 8:
            return "Incorrect compression method"
        if ((self["compression_info"].value << 12) +
            (self["compression_method"].value << 8) +
            (self["flag_compression_level"].value << 6) +
            (self["flag_dictionary_present"].value << 5) +
                (self["flag_check_bits"].value)) % 31 != 0:
            return "Invalid flag check value"
        return True

    def createFields(self):
        # CM
        yield Enum(Bits(self, "compression_method", 4), {8: "deflate", 15: "reserved"})
        # CINFO
        yield Bits(self, "compression_info", 4, "base-2 log of the window size")
        yield Bits(self, "flag_check_bits", 5)  # FCHECK
        yield Bit(self, "flag_dictionary_present")  # FDICT
        yield Enum(Bits(self, "flag_compression_level", 2),  # FLEVEL
                   {0: "Fastest", 1: "Fast", 2: "Default", 3: "Maximum, Slowest"})
        if self["flag_dictionary_present"].value:
            yield textHandler(UInt32(self, "dict_checksum", "ADLER32 checksum of dictionary information"), hexadecimal)
        yield DeflateData(self, "data", self.stream, description="Compressed Data")
        yield textHandler(UInt32(self, "data_checksum", "ADLER32 checksum of compressed data"), hexadecimal)


def zlib_inflate(stream, wbits=None):
    if wbits is None or wbits >= 0:
        return ZlibData(stream)["data"].uncompressed_data
    else:
        data = DeflateData(None, "root", stream, "", stream.askSize(None))
        for _ in data:
            pass
        return data.uncompressed_data



"""
LucasArts Font parser.

Author: Cyril Zorin
Creation date: 1 January 2007
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           UInt8, UInt16, UInt32, GenericVector)
from hachoir.core.endian import LITTLE_ENDIAN


class CharData(FieldSet):

    def __init__(self, chars, *args):
        FieldSet.__init__(self, *args)
        self.chars = chars

    def createFields(self):
        for char in self.chars:
            yield CharBitmap(char, self, "char_bitmap[]")


class CharBitmap(FieldSet):

    def __init__(self, char, *args):
        FieldSet.__init__(self, *args)
        self.char = char

    def createFields(self):
        width = self.char["width_pixels"].value
        for line in range(self.char["height_pixels"].value):
            yield GenericVector(self, "line[]", width,
                                UInt8, "pixel")


class CharInfo(FieldSet):
    static_size = 16 * 8

    def createFields(self):
        yield UInt32(self, "data_offset")
        yield UInt8(self, "logical_width")
        yield UInt8(self, "unknown[]")
        yield UInt8(self, "unknown[]")
        yield UInt8(self, "unknown[]")
        yield UInt32(self, "width_pixels")
        yield UInt32(self, "height_pixels")


class LafFile(Parser):
    PARSER_TAGS = {
        "id": "lucasarts_font",
        "category": "game",
        "file_ext": ("laf",),
        "min_size": 32 * 8,
        "description": "LucasArts Font"
    }

    endian = LITTLE_ENDIAN

    def validate(self):
        if self["num_chars"].value != 256:
            return "Invalid number of characters (%u)" % self["num_chars"].value
        if self["first_char_code"].value != 0:
            return "Invalid of code of first character code (%u)" % self["first_char_code"].value
        if self["last_char_code"].value != 255:
            return "Invalid of code of last character code (%u)" % self["last_char_code"].value
        if self["char_codes/char[0]"].value != 0:
            return "Invalid character code #0 (%u)" % self["char_codes/char[0]"].value
        if self["chars/char[0]/data_offset"].value != 0:
            return "Invalid character #0 offset"
        return True

    def createFields(self):
        yield UInt32(self, "num_chars")
        yield UInt32(self, "raw_font_data_size")
        yield UInt32(self, "max_char_width")
        yield UInt32(self, "min_char_width")
        yield UInt32(self, "unknown[]", 4)
        yield UInt32(self, "unknown[]", 4)
        yield UInt32(self, "first_char_code")
        yield UInt32(self, "last_char_code")

        yield GenericVector(self, "char_codes", self["num_chars"].value,
                            UInt16, "char")

        yield GenericVector(self, "chars", self["num_chars"].value,
                            CharInfo, "char")

        # character data. we make an effort to provide
        # something more meaningful than "RawBytes:
        # character bitmap data"
        yield CharData(self["chars"], self, "char_data")

        # read to the end
        if self.current_size < self._size:
            yield self.seekBit(self._size, "unknown[]")



"""
GIF picture parser.

Author: Victor Stinner, Robert Xiao

- GIF format
  http://local.wasp.uwa.edu.au/~pbourke/dataformats/gif/
- LZW compression
  http://en.wikipedia.org/wiki/LZW
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, ParserError,
                           Enum, UInt8, UInt16,
                           Bit, Bits, NullBytes,
                           String, PascalString8, Character,
                           NullBits, RawBytes,
                           CustomFragment)
from hachoir.parser.image.common import PaletteRGB
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.tools import humanDuration, paddingSize
from hachoir.core.text_handler import textHandler, displayHandler, hexadecimal

# Maximum image dimension (in pixel)
MAX_WIDTH = 6000
MAX_HEIGHT = MAX_WIDTH
MAX_FILE_SIZE = 100 * 1024 * 1024


def rle_repr(chain):
    """Run-length encode a list into an "eval"-able form

    Example:
    >>> rle_repr([20, 16, 16, 16, 16, 16, 18, 18, 65])
    '[20] + [16]*5 + [18]*2 + [65]'

    Adapted from http://twistedmatrix.com/trac/browser/trunk/twisted/python/dxprofile.py
    """
    def add_rle(previous, runlen, result):
        if isinstance(previous, (list, tuple)):
            previous = rle_repr(previous)
        if runlen > 1:
            result.append('[%s]*%i' % (previous, runlen))
        else:
            if result and '*' not in result[-1]:
                result[-1] = '[%s, %s]' % (result[-1][1:-1], previous)
            else:
                result.append('[%s]' % previous)
    iterable = iter(chain)
    runlen = 1
    result = []
    try:
        previous = next(iterable)
    except StopIteration:
        return "[]"
    for element in iterable:
        if element == previous:
            runlen = runlen + 1
            continue
        else:
            add_rle(previous, runlen, result)
            previous = element
            runlen = 1
    add_rle(previous, runlen, result)
    return ' + '.join(result)


class GifImageBlock(Parser):
    endian = LITTLE_ENDIAN

    def createFields(self):
        dictionary = {}
        self.nbits = self.startbits
        CLEAR_CODE = 2**self.nbits
        END_CODE = CLEAR_CODE + 1
        compress_code = CLEAR_CODE + 2
        obuf = []
        output = []
        while True:
            if compress_code >= 2**self.nbits:
                self.nbits += 1
            code = Bits(self, "code[]", self.nbits)
            if code.value == CLEAR_CODE:
                if compress_code == 2**(self.nbits - 1):
                    # this fixes a bizarre edge case where the reset code could
                    # appear just after the bits incremented. Apparently, the
                    # correct behaviour is to express the reset code with the
                    # old number of bits, not the new...
                    code = Bits(self, "code[]", self.nbits - 1)
                self.nbits = self.startbits + 1
                dictionary = {}
                compress_code = CLEAR_CODE + 2
                obuf = []
                code._description = "Reset Code (LZW code %i)" % code.value
                yield code
                continue
            elif code.value == END_CODE:
                code._description = "End of Information Code (LZW code %i)" % code.value
                yield code
                break
            if code.value < CLEAR_CODE:  # literal
                if obuf:
                    chain = obuf + [code.value]
                    dictionary[compress_code] = chain
                    compress_code += 1
                obuf = [code.value]
                output.append(code.value)
                code._description = "Literal Code %i" % code.value
            elif code.value >= CLEAR_CODE + 2:
                if code.value in dictionary:
                    chain = dictionary[code.value]
                    code._description = "Compression Code %i (found in dictionary as %s)" % (
                        code.value, rle_repr(chain))
                else:
                    chain = obuf + [obuf[0]]
                    code._description = "Compression Code %i (not found in dictionary; guessed to be %s)" % (
                        code.value, rle_repr(chain))
                dictionary[compress_code] = obuf + [chain[0]]
                compress_code += 1
                obuf = chain
                output += chain
            code._description += "; Current Decoded Length %i" % len(output)
            yield code
        padding = paddingSize(self.current_size, 8)
        if padding:
            yield NullBits(self, "padding[]", padding)


class Image(FieldSet):

    def createFields(self):
        yield UInt16(self, "left", "Left")
        yield UInt16(self, "top", "Top")
        yield UInt16(self, "width", "Width")
        yield UInt16(self, "height", "Height")

        yield Bits(self, "size_local_map", 3, "log2(size of local map) minus one")
        yield NullBits(self, "reserved", 2)
        yield Bit(self, "sort_flag", "Is the local map sorted by decreasing importance?")
        yield Bit(self, "interlaced", "Interlaced?")
        yield Bit(self, "has_local_map", "Use local color map?")

        if self["has_local_map"].value:
            nb_color = 1 << (1 + self["size_local_map"].value)
            yield PaletteRGB(self, "local_map", nb_color, "Local color map")

        yield UInt8(self, "lzw_min_code_size", "LZW Minimum Code Size")
        group = None
        while True:
            size = UInt8(self, "image_block_size[]")
            if size.value == 0:
                break
            yield size
            block = CustomFragment(
                self, "image_block[]", size.value * 8, GifImageBlock, "GIF Image Block", group)
            if group is None:
                block.group.args["startbits"] = self["lzw_min_code_size"].value
            group = block.group
            yield block
        yield NullBytes(self, "terminator", 1, "Terminator (0)")

    def createDescription(self):
        return "Image: %ux%u pixels at (%u,%u)" % (
            self["width"].value, self["height"].value,
            self["left"].value, self["top"].value)


DISPOSAL_METHOD = {
    0: "No disposal specified",
    1: "Do not dispose",
    2: "Restore to background color",
    3: "Restore to previous",
}

NETSCAPE_CODE = {
    1: "Loop count",
}


def parseApplicationExtension(parent):
    yield PascalString8(parent, "app_name", "Application name")
    while True:
        size = UInt8(parent, "size[]")
        if size.value == 0:
            break
        yield size
        if parent["app_name"].value == "NETSCAPE2.0" and size.value == 3:
            yield Enum(UInt8(parent, "netscape_code"), NETSCAPE_CODE)
            if parent["netscape_code"].value == 1:
                yield UInt16(parent, "loop_count")
            else:
                yield RawBytes(parent, "raw[]", 2)
        else:
            yield RawBytes(parent, "raw[]", size.value)
    yield NullBytes(parent, "terminator", 1, "Terminator (0)")


def parseGraphicControl(parent):
    yield UInt8(parent, "size", "Block size (4)")

    yield Bit(parent, "has_transp", "Has transparency")
    yield Bit(parent, "user_input", "User input")
    yield Enum(Bits(parent, "disposal_method", 3), DISPOSAL_METHOD)
    yield NullBits(parent, "reserved[]", 3)

    if parent["size"].value != 4:
        raise ParserError("Invalid graphic control size")
    yield displayHandler(UInt16(parent, "delay", "Delay time in millisecond"), humanDuration)
    yield UInt8(parent, "transp", "Transparent color index")
    yield NullBytes(parent, "terminator", 1, "Terminator (0)")


def parseComments(parent):
    while True:
        field = PascalString8(parent, "comment[]", strip=" \0\r\n\t")
        yield field
        if field.length == 0:
            break


def parseTextExtension(parent):
    yield UInt8(parent, "block_size", "Block Size")
    yield UInt16(parent, "left", "Text Grid Left")
    yield UInt16(parent, "top", "Text Grid Top")
    yield UInt16(parent, "width", "Text Grid Width")
    yield UInt16(parent, "height", "Text Grid Height")
    yield UInt8(parent, "cell_width", "Character Cell Width")
    yield UInt8(parent, "cell_height", "Character Cell Height")
    yield UInt8(parent, "fg_color", "Foreground Color Index")
    yield UInt8(parent, "bg_color", "Background Color Index")
    while True:
        field = PascalString8(parent, "comment[]", strip=" \0\r\n\t")
        yield field
        if field.length == 0:
            break


def defaultExtensionParser(parent):
    while True:
        size = UInt8(parent, "size[]", "Size (in bytes)")
        yield size
        if 0 < size.value:
            yield RawBytes(parent, "content[]", size.value)
        else:
            break


class Extension(FieldSet):
    ext_code = {
        0xf9: ("graphic_ctl[]", parseGraphicControl, "Graphic control"),
        0xfe: ("comments[]", parseComments, "Comments"),
        0xff: ("app_ext[]", parseApplicationExtension, "Application extension"),
        0x01: ("text_ext[]", parseTextExtension, "Plain text extension")
    }

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        code = self["code"].value
        if code in self.ext_code:
            self._name, self.parser, self._description = self.ext_code[code]
        else:
            self.parser = defaultExtensionParser

    def createFields(self):
        yield textHandler(UInt8(self, "code", "Extension code"), hexadecimal)
        yield from self.parser(self)

    def createDescription(self):
        return "Extension: function %s" % self["func"].display


class ScreenDescriptor(FieldSet):

    def createFields(self):
        yield UInt16(self, "width", "Width")
        yield UInt16(self, "height", "Height")
        yield Bits(self, "size_global_map", 3, "log2(size of global map) minus one")
        yield Bit(self, "sort_flag", "Is the global map sorted by decreasing importance?")
        yield Bits(self, "color_res", 3, "Color resolution minus one")
        yield Bit(self, "global_map", "Has global map?")
        yield UInt8(self, "background", "Background color")
        field = UInt8(self, "pixel_aspect_ratio")
        if field.value:
            field._description = "Pixel aspect ratio: %f (stored as %i)" % (
                (field.value + 15) / 64., field.value)
        else:
            field._description = "Pixel aspect ratio: not specified"
        yield field

    def createDescription(self):
        colors = 1 << (self["size_global_map"].value + 1)
        return "Screen descriptor: %ux%u pixels %u colors" \
            % (self["width"].value, self["height"].value, colors)


class GifFile(Parser):
    endian = LITTLE_ENDIAN
    separator_name = {
        "!": "Extension",
        ",": "Image",
        ";": "Terminator"
    }
    PARSER_TAGS = {
        "id": "gif",
        "category": "image",
        "file_ext": ("gif",),
        "mime": ("image/gif",),
        # signature + screen + separator + image
        "min_size": (6 + 7 + 1 + 9) * 8,
        "magic": ((b"GIF87a", 0), (b"GIF89a", 0)),
        "description": "GIF picture"
    }

    def validate(self):
        if self.stream.readBytes(0, 6) not in (b"GIF87a", b"GIF89a"):
            return "Wrong header"
        if self["screen/width"].value == 0 or self["screen/height"].value == 0:
            return "Invalid image size"
        if MAX_WIDTH < self["screen/width"].value:
            return "Image width too big (%u)" % self["screen/width"].value
        if MAX_HEIGHT < self["screen/height"].value:
            return "Image height too big (%u)" % self["screen/height"].value
        return True

    def createFields(self):
        # Header
        yield String(self, "magic", 3, "File magic code", charset="ASCII")
        yield String(self, "version", 3, "GIF version", charset="ASCII")

        yield ScreenDescriptor(self, "screen")
        if self["screen/global_map"].value:
            bpp = (self["screen/size_global_map"].value + 1)
            yield PaletteRGB(self, "color_map", 1 << bpp, "Color map")
            self.color_map = self["color_map"]
        else:
            self.color_map = None

        self.images = []
        while True:
            code = Enum(Character(self, "separator[]",
                                  "Separator code"), self.separator_name)
            yield code
            code = code.value
            if code == "!":
                yield Extension(self, "extensions[]")
            elif code == ",":
                yield Image(self, "image[]")
            elif code == ";":
                # GIF Terminator
                break
            else:
                raise ParserError(
                    "Wrong GIF image separator: 0x%02X" % ord(code))

    def createContentSize(self):
        field = self["image[0]"]
        start = field.absolute_address + field.size
        end = start + MAX_FILE_SIZE * 8
        pos = self.stream.searchBytes(b"\0;", start, end)
        if pos:
            return pos + 16
        return None



from hachoir.core.tools import makeUnicode, normalizeNewline
from hachoir.metadata import config
from hachoir.metadata.setter import normalizeString

MIN_PRIORITY = 100
MAX_PRIORITY = 999

QUALITY_FASTEST = 0.0
QUALITY_FAST = 0.25
QUALITY_NORMAL = 0.5
QUALITY_GOOD = 0.75
QUALITY_BEST = 1.0


class DataValue:

    def __init__(self, value, text):
        self.value = value
        self.text = text


class Data:

    def __init__(self, key, priority, description,
                 text_handler=None, type=None, filter=None, conversion=None):
        """
        handler is only used if value is not string nor unicode, prototype:
           def handler(value) -> str/unicode
        """
        assert MIN_PRIORITY <= priority <= MAX_PRIORITY
        assert isinstance(description, str)
        self.metadata = None
        self.key = key
        self.description = description
        self.values = []
        if type and not isinstance(type, (tuple, list)):
            type = (type,)
        self.type = type
        self.text_handler = text_handler
        self.filter = filter
        self.priority = priority
        self.conversion = conversion

    def __lt__(self, other):
        return self.priority < other.priority

    def _createItem(self, value, text=None):
        if text is None:
            if isinstance(value, str):
                text = value
            elif self.text_handler:
                text = self.text_handler(value)
                assert isinstance(text, str)
            else:
                text = makeUnicode(value)
        return DataValue(value, text)

    def add(self, value):
        if isinstance(value, tuple):
            if len(value) != 2:
                raise ValueError("Data.add() only accept "
                                 "tuple of 2 elements: (value,text)")
            value, text = value
        else:
            text = None

        # Skip value 'None'
        if value is None:
            return

        # Convert string to Unicode string using charset ISO-8859-1
        if self.conversion:
            try:
                new_value = self.conversion(self.metadata, self.key, value)
            except Exception as err:
                self.metadata.warning("Error during conversion of %r value: %s"
                                      % (self.key, err))
                return
            if new_value is None:
                dest_types = " or ".join(str(item.__name__)
                                         for item in self.type)
                self.metadata.warning("Unable to convert %s=%r (%s) to %s" % (
                    self.key, value, type(value).__name__, dest_types))
                return
            if isinstance(new_value, tuple):
                if text:
                    value = new_value[0]
                else:
                    value, text = new_value
            else:
                value = new_value
        elif isinstance(value, bytes):
            value = str(value, "ISO-8859-1")

        if isinstance(value, str):
            value = normalizeString(value)
            if not value:
                return

        if self.type and not isinstance(value, self.type):
            dest_types = " or ".join(str(item.__name__) for item in self.type)
            self.metadata.warning("Key %r: value %r type (%s) is not %s" % (
                self.key, value, type(value).__name__, dest_types))
            return

        # Skip empty strings
        if isinstance(value, str):
            value = normalizeNewline(value)
            if (config.MAX_STR_LENGTH
                    and config.MAX_STR_LENGTH < len(value)):
                value = value[:config.MAX_STR_LENGTH] + "(...)"

        # Skip duplicates
        if value in self:
            return

        # Use filter
        if self.filter and not self.filter(value):
            self.metadata.warning("Skip value %s=%r (filter)"
                                  % (self.key, value))
            return

        # For string, if you have "verlongtext" and "verylo",
        # keep the longer value
        if isinstance(value, str):
            for index, item in enumerate(self.values):
                item = item.value
                if not isinstance(item, str):
                    continue
                if value.startswith(item):
                    # Find longer value, replace the old one
                    self.values[index] = self._createItem(value, text)
                    return
                if item.startswith(value):
                    # Find truncated value, skip it
                    return

        # Add new value
        self.values.append(self._createItem(value, text))

    def __len__(self):
        return len(self.values)

    def __getitem__(self, index):
        return self.values[index]

    def __contains__(self, value):
        for item in self.values:
            if value == item.value:
                return True
        return False



from hachoir.field import Bytes
from hachoir.core.tools import makePrintable, humanFilesize
from hachoir.stream import InputIOStream


class SubFile(Bytes):
    """
    File stored in another file
    """

    def __init__(self, parent, name, length, description=None,
                 parser=None, filename=None, mime_type=None, parser_class=None):
        if filename:
            if not isinstance(filename, str):
                filename = makePrintable(filename, "ISO-8859-1")
            if not description:
                description = 'File "%s" (%s)' % (
                    filename, humanFilesize(length))
        Bytes.__init__(self, parent, name, length, description)

        def createInputStream(cis, **args):
            tags = args.setdefault("tags", [])
            if parser_class:
                tags.append(("class", parser_class))
            if parser is not None:
                tags.append(("id", parser.PARSER_TAGS["id"]))
            if mime_type:
                tags.append(("mime", mime_type))
            if filename:
                tags.append(("filename", filename))
            return cis(**args)
        self.setSubIStream(createInputStream)


class CompressedStream:
    offset = 0

    def __init__(self, stream, decompressor):
        self.stream = stream
        self.decompressor = decompressor(stream)
        self._buffer = b''

    def read(self, size):
        d = self._buffer
        data = [d[:size]]
        size -= len(d)
        if size > 0:
            d = self.decompressor(size)
            data.append(d[:size])
            size -= len(d)
            while size > 0:
                n = 4096
                if self.stream.size:
                    n = min(self.stream.size - self.offset, n)
                    if not n:
                        break
                d = self.stream.read(self.offset, n)[1]
                self.offset += 8 * len(d)
                d = self.decompressor(size, d)
                data.append(d[:size])
                size -= len(d)
        self._buffer = d[size + len(d):]
        return b''.join(data)


def CompressedField(field, decompressor):
    def createInputStream(cis, source=None, **args):
        if field._parent:
            stream = cis(source=source)
            args.setdefault("tags", []).extend(stream.tags)
        else:
            stream = field.stream
        input = CompressedStream(stream, decompressor)
        if source is None:
            source = "Compressed source: '%s' (offset=%s)" % (
                stream.source, field.absolute_address)
        return InputIOStream(input, source=source, **args)
    field.setSubIStream(createInputStream)
    return field



"""
Parser list managment:
- createParser() find the best parser for a file.
"""

import os
from hachoir.core.error import warning, info
from hachoir.parser import ValidateError, HachoirParserList
from hachoir.stream import FileInputStream
import weakref


class QueryParser(object):
    fallback = None
    other = None

    def __init__(self, tags):
        self.validate = True
        self.use_fallback = False
        self.parser_args = None
        self.db = HachoirParserList.getInstance()
        self.parsers = set(self.db)
        parsers = []
        for tag in tags:
            if not self.parsers:
                break
            parsers += self._getByTag(tag)
            if self.fallback is None:
                self.fallback = len(parsers) == 1
        if self.parsers:
            other = len(parsers)
            parsers += list(self.parsers)
            self.other = parsers[other]
        self.parsers = parsers

    def __iter__(self):
        return iter(self.parsers)

    def translate(self, name, value):
        if name == "filename":
            filename = os.path.basename(value).split(".")
            if len(filename) <= 1:
                value = ""
            else:
                value = filename[-1].lower()
            name = "file_ext"
        return name, value

    def _getByTag(self, tag):
        if tag is None:
            self.parsers.clear()
            return []
        elif callable(tag):
            parsers = [parser for parser in self.parsers if tag(parser)]
            for parser in parsers:
                self.parsers.remove(parser)
        elif tag[0] == "class":
            self.validate = False
            return [tag[1]]
        elif tag[0] == "args":
            self.parser_args = tag[1]
            return []
        else:
            tag = self.translate(*tag)
            parsers = []
            if tag is not None:
                key = tag[0]
                byname = self.db.bytag.get(key, {})
                if tag[1] is None:
                    values = iter(byname.values())
                else:
                    values = byname.get(tag[1], ()),
                if key == "id" and values:
                    self.validate = False
                for value in values:
                    for parser in value:
                        if parser in self.parsers:
                            parsers.append(parser)
                            self.parsers.remove(parser)
        return parsers

    def parse(self, stream, fallback=True):
        if hasattr(stream, "_cached_parser"):
            parser = stream._cached_parser()
        else:
            parser = None
        if parser is not None:
            if parser.__class__ in self.parsers:
                return parser
        parser = self.doparse(stream, fallback)
        if parser is not None:
            stream._cached_parser = weakref.ref(parser)
        return parser

    def doparse(self, stream, fallback=True):
        fb = None
        warn = warning
        for parser in self.parsers:
            try:
                parser_obj = parser(stream, validate=self.validate)
                if self.parser_args:
                    for key, value in self.parser_args.items():
                        setattr(parser_obj, key, value)
                return parser_obj
            except ValidateError as err:
                if fallback and self.fallback:
                    fb = parser
                if parser == self.other:
                    warn = info
                warn("Skip parser '%s': %s" % (parser.__name__, err))
            except Exception as err:
                if parser == self.other:
                    warn = info
                warn("Skip parser '%s': %s" % (parser.__name__, err))
            fallback = False
        if self.use_fallback and fb:
            warning("Force use of parser '%s'" % fb.__name__)
            return fb(stream)


def guessParser(stream):
    return QueryParser(stream.tags).parse(stream)


def createParser(filename, real_filename=None, tags=None):
    """
    Create a parser from a file or returns None on error.

    Options:
    - file (str|io.IOBase): Input file name or
        a byte io.IOBase stream  ;
    - real_filename (str): Real file name.
    """
    if not tags:
        tags = []
    stream = FileInputStream(filename, real_filename, tags=tags)
    guess = guessParser(stream)
    if guess is None:
        stream.close()
    return guess



"""
ms task/job file parser

Author: Jeff Bryner
Creation date: 2010-11
References:
http://msdn.microsoft.com/en-us/library/cc248286%28v=PROT.13%29.aspx
http://msdn.microsoft.com/en-us/library/cc248287%28v=PROT.13%29.aspx
http://technet.microsoft.com/en-us/library/bb490996.aspx
"""


from hachoir.parser import Parser
from hachoir.field import (FieldSet, RootSeekableFieldSet,
                           UInt32, UInt16,
                           Bit, RawBits, RawBytes, Enum)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.parser.common.win32 import PascalStringWin16, GUID


class TaskTrigger(FieldSet):
    TRIGGER_TYPE = {
        0x00000000: "ONCE",
        0x00000001: "DAILY",
        0x00000002: "WEEKLY",
        0x00000003: "MONTHLYDATE",
        0x00000004: "MONTHLYDOW",
        0x00000005: "EVENT_ON_IDLE",
        0x00000006: "EVENT_AT_SYSTEMSTART",
        0x00000007: "EVENT_AT_LOGON"
    }

    def __init__(self, *args, **kwargs):
        FieldSet.__init__(self, *args, **kwargs)
        self._size = self["TriggerSize"].value * 8

    def createFields(self):
        yield UInt16(self, "TriggerSize")
        yield UInt16(self, "Reserved[]")
        yield UInt16(self, "BeginYear")
        yield UInt16(self, "BeginMonth")
        yield UInt16(self, "BeginDay")
        yield UInt16(self, "EndYear")
        yield UInt16(self, "EndMonth")
        yield UInt16(self, "EndDay")
        yield UInt16(self, "StartHour")
        yield UInt16(self, "StartMinute")
        yield UInt32(self, "MinutesDuration")
        yield UInt32(self, "MinutesInterval", "Time period between repeated trigger firings.")
        yield Bit(self, "HasEndDate", "Can task stop at some point in time?")
        yield Bit(self, "KillAtDurationEnd", "Can task be stopped at the end of the repetition period?")
        yield Bit(self, "TriggerDisabled", "Is this trigger disabled?")
        yield RawBits(self, "Unused[]", 29)
        yield Enum(UInt32(self, "TriggerType"), self.TRIGGER_TYPE)
        yield UInt16(self, "TriggerSpecific0")
        yield UInt16(self, "TriggerSpecific1")
        yield UInt16(self, "TriggerSpecific2")
        yield UInt16(self, "Padding")
        yield UInt16(self, "Reserved[]")
        yield UInt16(self, "Reserved[]")


class MSTaskFile(Parser, RootSeekableFieldSet):
    PARSER_TAGS = {
        "id": "mstask",
        "category": "misc",    # "archive", "audio", "container", ...
        # TODO: Example ("bmp",) to parse the file "image.bmp"
        "file_ext": ("job",),
        # TODO: Minimum file size (x bits, or x*8 in bytes)
        "min_size": 100,
        # TODO: Example: "A bitmap picture",
        "description": ".job 'at' file parser from ms windows",
    }

    endian = LITTLE_ENDIAN

    PRODUCT_VERSION = {
        0x0400: "Windows NT 4.0",
        0x0500: "Windows 2000",
        0x0501: "Windows XP",
        0x0600: "Windows Vista",
        0x0601: "Windows 7"
    }

    TASK_STATUS = {
        0x00041300: "Task Ready",
        0x00041301: "Task running",
        0x00041302: "Task disabled",
        0x00041303: "Task has not run",
        0x00041304: "Task has no more runs",
        0x00041305: "Task not scheduled",
        0x00041306: "Task terminated",
        0x00041307: "Task has no valid triggers",
        0x00041308: "Task contains only event triggers that do not have set run times",
        0x00041309: "Task trigger not found",
        0x0004130A: "One or more of the properties that are required to run this task have not been set.",
        0x0004130B: "There is no running instance of the task",
        0x0004130C: "Task Schedule Remoting Protocol service is not installed",
        0x0004130D: "Task object cannot be opened",
        0x0004130E: "Task object is invalid",
        0x0004130F: "No Account information could be found in Task Scheduler Remoting Protocol security database for the task indicated."
    }

    def validate(self):
        # The MAGIC for a task file is the windows version that created it
        # http://msdn.microsoft.com/en-us/library/2d1fbbab-fe6c-4ae5-bdf5-41dc526b2439%28v=PROT.13%29#id11
        if self['WindowsVersion'].value not in self.PRODUCT_VERSION:
            return "Invalid Product Version Field"
        return True

    def createFields(self):
        yield Enum(UInt16(self, "WindowsVersion"), self.PRODUCT_VERSION)
        yield UInt16(self, "FileVersion")
        yield GUID(self, "JobUUID")
        yield UInt16(self, "AppNameOffset", "App Name Length Offset")
        yield UInt16(self, "TriggerOffset", "Contains the offset in bytes within the .JOB file where the task triggers are located.")
        yield UInt16(self, "ErrorRetryCount", "Contains the number of execute attempts that are attempted for the task if the task fails to start.")
        yield UInt16(self, "ErrorRetryInterval", "Contains the interval, in minutes, between successive retries")
        yield UInt16(self, "IdleDeadline", "Contains a maximum time in minutes to wait for the machine to become idle for Idle Wait minutes.")
        yield UInt16(self, "IdleWait", "Contains a value in minutes. The machine remains idle for this many minutes before it runs the task")
        yield UInt32(self, "Priority")
        yield UInt32(self, "MaxRunTime", "Maximum run time in milliseconds")
        yield UInt32(self, "ExitCode", "This contains the exit code of the executed task upon the completion of that task.")
        yield Enum(UInt32(self, "Status"), self.TASK_STATUS)
        yield Bit(self, "Interactive", "Can Task interact with user?")
        yield Bit(self, "DeleteWhenDone", "Remove the task file when done?")
        yield Bit(self, "Disabled", "Is Task disabled?")
        yield Bit(self, "StartOnlyIfIdle", "Task begins only if computer is not in use at the scheduled time")
        yield Bit(self, "KillOnIdleEnd", "Kill task if user input is detected, terminating idle state?")
        yield Bit(self, "DontStartIfOnBatteries")
        yield Bit(self, "KillIfGoingOnBatteries")
        yield Bit(self, "RunOnlyIfDocked")
        yield Bit(self, "HiddenTask")
        yield Bit(self, "RunIfConnectedToInternet")
        yield Bit(self, "RestartOnIdleResume")
        yield Bit(self, "SystemRequired", "Can task cause system to resume or awaken if system is sleeping?")
        yield Bit(self, "OnlyIfUserLoggedOn")
        yield Bit(self, "ApplicationNameExists", "Does task have an application name defined?")
        yield Bit(self, "Unused[]")
        yield Bit(self, "Unused[]")
        yield RawBytes(self, "flags", 2)
        yield UInt16(self, "LastRunYear")
        yield UInt16(self, "LastRunMonth")
        yield UInt16(self, "LastRunWeekday", "Sunday=0,Saturday=6")
        yield UInt16(self, "LastRunDay")
        yield UInt16(self, "LastRunHour")
        yield UInt16(self, "LastRunMinute")
        yield UInt16(self, "LastRunSecond")
        yield UInt16(self, "LastRunMillisecond")
        yield UInt16(self, "RunningInstanceCount")
        yield PascalStringWin16(self, "AppNameLength", strip='\0')
        yield PascalStringWin16(self, "Parameters", strip='\0')
        yield PascalStringWin16(self, "WorkingDirectory", strip='\0')
        yield PascalStringWin16(self, "Author", strip='\0')
        yield PascalStringWin16(self, "Comment", strip='\0')

        yield UInt16(self, "UserDataSize")
        # todo: read optional userdata
        yield UInt16(self, "ReservedDataSize")
        if self["ReservedDataSize"].value == 8:
            yield Enum(UInt32(self, "StartError", "contains the HRESULT error from the most recent attempt to start the task"), self.TASK_STATUS)
            yield UInt32(self, "TaskFlags")
        elif self["ReservedDataSize"].value:
            yield RawBytes(self, "Reserved", self["ReservedDataSize"].value)
        yield UInt16(self, "TriggerCount", "size of the array of triggers")
        for i in range(self["TriggerCount"].value):
            yield TaskTrigger(self, "Trigger[]")



from hachoir.parser.image.bmp import BmpFile  # noqa
from hachoir.parser.image.gif import GifFile  # noqa
from hachoir.parser.image.ico import IcoFile  # noqa
from hachoir.parser.image.jpeg import JpegFile  # noqa
from hachoir.parser.image.pcx import PcxFile  # noqa
from hachoir.parser.image.psd import PsdFile  # noqa
from hachoir.parser.image.png import PngFile  # noqa
from hachoir.parser.image.tga import TargaFile  # noqa
from hachoir.parser.image.tiff import TiffFile  # noqa
from hachoir.parser.image.wmf import WMF_File  # noqa
from hachoir.parser.image.xcf import XcfFile  # noqa
from hachoir.parser.image.cr2 import CR2File  # noqa



from hachoir.field import joinPath, MissingField


class EditorError(Exception):
    pass


class FakeField(object):
    """
    This class have API looks similar to Field API, but objects don't contain
    any value: all values are _computed_ by parent methods.

    Example: FakeField(editor, "abc").size calls editor._getFieldSize("abc").
    """
    is_field_set = False

    def __init__(self, parent, name):
        self._parent = parent
        self._name = name

    def _getPath(self):
        return joinPath(self._parent.path, self._name)
    path = property(_getPath)

    def _getName(self):
        return self._name
    name = property(_getName)

    def _getAddress(self):
        return self._parent._getFieldAddress(self._name)
    address = property(_getAddress)

    def _getSize(self):
        return self._parent.input[self._name].size
    size = property(_getSize)

    def _getValue(self):
        return self._parent.input[self._name].value
    value = property(_getValue)

    def createDisplay(self):
        # TODO: Returns new value if field is altered
        return self._parent.input[self._name].display
    display = property(createDisplay)

    def _getParent(self):
        return self._parent
    parent = property(_getParent)

    def hasValue(self):
        return self._parent.input[self._name].hasValue()

    def __getitem__(self, key):
        # TODO: Implement this function!
        raise MissingField(self, key)

    def _isAltered(self):
        return False
    is_altered = property(_isAltered)

    def writeInto(self, output):
        size = self.size
        addr = self._parent._getFieldInputAddress(self._name)
        input = self._parent.input
        stream = input.stream
        if size % 8 or addr % 8:
            output.copyBitsFrom(stream, addr, size, input.endian)
        else:
            output.copyBytesFrom(stream, addr, size // 8)



"""MAR (Mozilla ARchive) parser

Author: Robert Xiao
Creation date: July 10, 2007

"""

from hachoir.core.endian import BIG_ENDIAN
from hachoir.field import (RootSeekableFieldSet, FieldSet,
                           String, CString, UInt32, RawBytes)
from hachoir.core.text_handler import displayHandler, filesizeHandler
from hachoir.core.tools import humanUnixAttributes
from hachoir.parser import HachoirParser


class IndexEntry(FieldSet):

    def createFields(self):
        yield UInt32(self, "offset", "Offset in bytes relative to start of archive")
        yield filesizeHandler(UInt32(self, "length", "Length in bytes"))
        yield displayHandler(UInt32(self, "flags"), humanUnixAttributes)
        yield CString(self, "name", "Filename (byte array)")

    def createDescription(self):
        return 'File %s, Size %s, Mode %s' % (
            self["name"].display, self["length"].display, self["flags"].display)


class MozillaArchive(HachoirParser, RootSeekableFieldSet):
    MAGIC = b"MAR1"
    PARSER_TAGS = {
        "id": "mozilla_ar",
        "category": "archive",
        "file_ext": ("mar",),
        "min_size": (8 + 4 + 13) * 8,  # Header, Index Header, 1 Index Entry
        "magic": ((MAGIC, 0),),
        "description": "Mozilla Archive",
    }
    endian = BIG_ENDIAN

    def __init__(self, stream, **args):
        RootSeekableFieldSet.__init__(
            self, None, "root", stream, None, stream.askSize(self))
        HachoirParser.__init__(self, stream, **args)

    def validate(self):
        if self.stream.readBytes(0, 4) != self.MAGIC:
            return "Invalid magic"
        return True

    def createFields(self):
        yield String(self, "magic", 4, "File signature (MAR1)", charset="ASCII")
        yield UInt32(self, "index_offset", "Offset to index relative to file start")
        self.seekByte(self["index_offset"].value, False)
        yield UInt32(self, "index_size", "size of index in bytes")
        current_index_size = 0  # bytes
        while current_index_size < self["index_size"].value:
            # plus 4 compensates for index_size
            self.seekByte(self["index_offset"].value +
                          current_index_size + 4, False)
            entry = IndexEntry(self, "index_entry[]")
            yield entry
            current_index_size += entry.size // 8
            self.seekByte(entry["offset"].value, False)
            yield RawBytes(self, "file[]", entry["length"].value)



"""
ISO639-2 standart: the module only contains the dictionary ISO639_2
which maps a language code in three letters (eg. "fre") to a language
name in english (eg. "French").
"""

# ISO-639, the list comes from:
# http://www.loc.gov/standards/iso639-2/php/English_list.php
_ISO639 = (
    ("Abkhazian", "abk", "ab"),
    ("Achinese", "ace", None),
    ("Acoli", "ach", None),
    ("Adangme", "ada", None),
    ("Adygei", "ady", None),
    ("Adyghe", "ady", None),
    ("Afar", "aar", "aa"),
    ("Afrihili", "afh", None),
    ("Afrikaans", "afr", "af"),
    ("Afro-Asiatic (Other)", "afa", None),
    ("Ainu", "ain", None),
    ("Akan", "aka", "ak"),
    ("Akkadian", "akk", None),
    ("Albanian", "alb/sqi", "sq"),
    ("Alemani", "gsw", None),
    ("Aleut", "ale", None),
    ("Algonquian languages", "alg", None),
    ("Altaic (Other)", "tut", None),
    ("Amharic", "amh", "am"),
    ("Angika", "anp", None),
    ("Apache languages", "apa", None),
    ("Arabic", "ara", "ar"),
    ("Aragonese", "arg", "an"),
    ("Aramaic", "arc", None),
    ("Arapaho", "arp", None),
    ("Araucanian", "arn", None),
    ("Arawak", "arw", None),
    ("Armenian", "arm/hye", "hy"),
    ("Aromanian", "rup", None),
    ("Artificial (Other)", "art", None),
    ("Arumanian", "rup", None),
    ("Assamese", "asm", "as"),
    ("Asturian", "ast", None),
    ("Athapascan languages", "ath", None),
    ("Australian languages", "aus", None),
    ("Austronesian (Other)", "map", None),
    ("Avaric", "ava", "av"),
    ("Avestan", "ave", "ae"),
    ("Awadhi", "awa", None),
    ("Aymara", "aym", "ay"),
    ("Azerbaijani", "aze", "az"),
    ("Bable", "ast", None),
    ("Balinese", "ban", None),
    ("Baltic (Other)", "bat", None),
    ("Baluchi", "bal", None),
    ("Bambara", "bam", "bm"),
    ("Bamileke languages", "bai", None),
    ("Banda", "bad", None),
    ("Bantu (Other)", "bnt", None),
    ("Basa", "bas", None),
    ("Bashkir", "bak", "ba"),
    ("Basque", "baq/eus", "eu"),
    ("Batak (Indonesia)", "btk", None),
    ("Beja", "bej", None),
    ("Belarusian", "bel", "be"),
    ("Bemba", "bem", None),
    ("Bengali", "ben", "bn"),
    ("Berber (Other)", "ber", None),
    ("Bhojpuri", "bho", None),
    ("Bihari", "bih", "bh"),
    ("Bikol", "bik", None),
    ("Bilin", "byn", None),
    ("Bini", "bin", None),
    ("Bislama", "bis", "bi"),
    ("Blin", "byn", None),
    ("Bokmål, Norwegian", "nob", "nb"),
    ("Bosnian", "bos", "bs"),
    ("Braj", "bra", None),
    ("Breton", "bre", "br"),
    ("Buginese", "bug", None),
    ("Bulgarian", "bul", "bg"),
    ("Buriat", "bua", None),
    ("Burmese", "bur/mya", "my"),
    ("Caddo", "cad", None),
    ("Carib", "car", None),
    ("Castilian", "spa", "es"),
    ("Catalan", "cat", "ca"),
    ("Caucasian (Other)", "cau", None),
    ("Cebuano", "ceb", None),
    ("Celtic (Other)", "cel", None),
    ("Central American Indian (Other)", "cai", None),
    ("Chagatai", "chg", None),
    ("Chamic languages", "cmc", None),
    ("Chamorro", "cha", "ch"),
    ("Chechen", "che", "ce"),
    ("Cherokee", "chr", None),
    ("Chewa", "nya", "ny"),
    ("Cheyenne", "chy", None),
    ("Chibcha", "chb", None),
    ("Chichewa", "nya", "ny"),
    ("Chinese", "chi/zho", "zh"),
    ("Chinook jargon", "chn", None),
    ("Chipewyan", "chp", None),
    ("Choctaw", "cho", None),
    ("Chuang", "zha", "za"),
    ("Church Slavic", "chu", "cu"),
    ("Church Slavonic", "chu", "cu"),
    ("Chuukese", "chk", None),
    ("Chuvash", "chv", "cv"),
    ("Classical Nepal Bhasa", "nwc", None),
    ("Classical Newari", "nwc", None),
    ("Coptic", "cop", None),
    ("Cornish", "cor", "kw"),
    ("Corsican", "cos", "co"),
    ("Cree", "cre", "cr"),
    ("Creek", "mus", None),
    ("Creoles and pidgins (Other)", "crp", None),
    ("Creoles and pidgins, English based (Other)", "cpe", None),
    ("Creoles and pidgins, French-based (Other)", "cpf", None),
    ("Creoles and pidgins, Portuguese-based (Other)", "cpp", None),
    ("Crimean Tatar", "crh", None),
    ("Crimean Turkish", "crh", None),
    ("Croatian", "scr/hrv", "hr"),
    ("Cushitic (Other)", "cus", None),
    ("Czech", "cze/ces", "cs"),
    ("Dakota", "dak", None),
    ("Danish", "dan", "da"),
    ("Dargwa", "dar", None),
    ("Dayak", "day", None),
    ("Delaware", "del", None),
    ("Dhivehi", "div", "dv"),
    ("Dimili", "zza", None),
    ("Dimli", "zza", None),
    ("Dinka", "din", None),
    ("Divehi", "div", "dv"),
    ("Dogri", "doi", None),
    ("Dogrib", "dgr", None),
    ("Dravidian (Other)", "dra", None),
    ("Duala", "dua", None),
    ("Dutch", "dut/nld", "nl"),
    ("Dutch, Middle (ca.1050-1350)", "dum", None),
    ("Dyula", "dyu", None),
    ("Dzongkha", "dzo", "dz"),
    ("Eastern Frisian", "frs", None),
    ("Efik", "efi", None),
    ("Egyptian (Ancient)", "egy", None),
    ("Ekajuk", "eka", None),
    ("Elamite", "elx", None),
    ("English", "eng", "en"),
    ("English, Middle (1100-1500)", "enm", None),
    ("English, Old (ca.450-1100)", "ang", None),
    ("Erzya", "myv", None),
    ("Esperanto", "epo", "eo"),
    ("Estonian", "est", "et"),
    ("Ewe", "ewe", "ee"),
    ("Ewondo", "ewo", None),
    ("Fang", "fan", None),
    ("Fanti", "fat", None),
    ("Faroese", "fao", "fo"),
    ("Fijian", "fij", "fj"),
    ("Filipino", "fil", None),
    ("Finnish", "fin", "fi"),
    ("Finno-Ugrian (Other)", "fiu", None),
    ("Flemish", "dut/nld", "nl"),
    ("Fon", "fon", None),
    ("French", "fre/fra", "fr"),
    ("French, Middle (ca.1400-1600)", "frm", None),
    ("French, Old (842-ca.1400)", "fro", None),
    ("Friulian", "fur", None),
    ("Fulah", "ful", "ff"),
    ("Ga", "gaa", None),
    ("Gaelic", "gla", "gd"),
    ("Galician", "glg", "gl"),
    ("Ganda", "lug", "lg"),
    ("Gayo", "gay", None),
    ("Gbaya", "gba", None),
    ("Geez", "gez", None),
    ("Georgian", "geo/kat", "ka"),
    ("German", "ger/deu", "de"),
    ("German, Low", "nds", None),
    ("German, Middle High (ca.1050-1500)", "gmh", None),
    ("German, Old High (ca.750-1050)", "goh", None),
    ("Germanic (Other)", "gem", None),
    ("Gikuyu", "kik", "ki"),
    ("Gilbertese", "gil", None),
    ("Gondi", "gon", None),
    ("Gorontalo", "gor", None),
    ("Gothic", "got", None),
    ("Grebo", "grb", None),
    ("Greek, Ancient (to 1453)", "grc", None),
    ("Greek, Modern (1453-)", "gre/ell", "el"),
    ("Greenlandic", "kal", "kl"),
    ("Guarani", "grn", "gn"),
    ("Gujarati", "guj", "gu"),
    ("Gwich´in", "gwi", None),
    ("Haida", "hai", None),
    ("Haitian", "hat", "ht"),
    ("Haitian Creole", "hat", "ht"),
    ("Hausa", "hau", "ha"),
    ("Hawaiian", "haw", None),
    ("Hebrew", "heb", "he"),
    ("Herero", "her", "hz"),
    ("Hiligaynon", "hil", None),
    ("Himachali", "him", None),
    ("Hindi", "hin", "hi"),
    ("Hiri Motu", "hmo", "ho"),
    ("Hittite", "hit", None),
    ("Hmong", "hmn", None),
    ("Hungarian", "hun", "hu"),
    ("Hupa", "hup", None),
    ("Iban", "iba", None),
    ("Icelandic", "ice/isl", "is"),
    ("Ido", "ido", "io"),
    ("Igbo", "ibo", "ig"),
    ("Ijo", "ijo", None),
    ("Iloko", "ilo", None),
    ("Inari Sami", "smn", None),
    ("Indic (Other)", "inc", None),
    ("Indo-European (Other)", "ine", None),
    ("Indonesian", "ind", "id"),
    ("Ingush", "inh", None),
    ("Interlingua", "ina", "ia"),
    ("Interlingue", "ile", "ie"),
    ("Inuktitut", "iku", "iu"),
    ("Inupiaq", "ipk", "ik"),
    ("Iranian (Other)", "ira", None),
    ("Irish", "gle", "ga"),
    ("Irish, Middle (900-1200)", "mga", None),
    ("Irish, Old (to 900)", "sga", None),
    ("Iroquoian languages", "iro", None),
    ("Italian", "ita", "it"),
    ("Japanese", "jpn", "ja"),
    ("Javanese", "jav", "jv"),
    ("Judeo-Arabic", "jrb", None),
    ("Judeo-Persian", "jpr", None),
    ("Kabardian", "kbd", None),
    ("Kabyle", "kab", None),
    ("Kachin", "kac", None),
    ("Kalaallisut", "kal", "kl"),
    ("Kalmyk", "xal", None),
    ("Kamba", "kam", None),
    ("Kannada", "kan", "kn"),
    ("Kanuri", "kau", "kr"),
    ("Kara-Kalpak", "kaa", None),
    ("Karachay-Balkar", "krc", None),
    ("Karelian", "krl", None),
    ("Karen", "kar", None),
    ("Kashmiri", "kas", "ks"),
    ("Kashubian", "csb", None),
    ("Kawi", "kaw", None),
    ("Kazakh", "kaz", "kk"),
    ("Khasi", "kha", None),
    ("Khmer", "khm", "km"),
    ("Khoisan (Other)", "khi", None),
    ("Khotanese", "kho", None),
    ("Kikuyu", "kik", "ki"),
    ("Kimbundu", "kmb", None),
    ("Kinyarwanda", "kin", "rw"),
    ("Kirdki", "zza", None),
    ("Kirghiz", "kir", "ky"),
    ("Kirmanjki", "zza", None),
    ("Klingon", "tlh", None),
    ("Komi", "kom", "kv"),
    ("Kongo", "kon", "kg"),
    ("Konkani", "kok", None),
    ("Korean", "kor", "ko"),
    ("Kosraean", "kos", None),
    ("Kpelle", "kpe", None),
    ("Kru", "kro", None),
    ("Kuanyama", "kua", "kj"),
    ("Kumyk", "kum", None),
    ("Kurdish", "kur", "ku"),
    ("Kurukh", "kru", None),
    ("Kutenai", "kut", None),
    ("Kwanyama", "kua", "kj"),
    ("Ladino", "lad", None),
    ("Lahnda", "lah", None),
    ("Lamba", "lam", None),
    ("Lao", "lao", "lo"),
    ("Latin", "lat", "la"),
    ("Latvian", "lav", "lv"),
    ("Letzeburgesch", "ltz", "lb"),
    ("Lezghian", "lez", None),
    ("Limburgan", "lim", "li"),
    ("Limburger", "lim", "li"),
    ("Limburgish", "lim", "li"),
    ("Lingala", "lin", "ln"),
    ("Lithuanian", "lit", "lt"),
    ("Lojban", "jbo", None),
    ("Low German", "nds", None),
    ("Low Saxon", "nds", None),
    ("Lower Sorbian", "dsb", None),
    ("Lozi", "loz", None),
    ("Luba-Katanga", "lub", "lu"),
    ("Luba-Lulua", "lua", None),
    ("Luiseno", "lui", None),
    ("Lule Sami", "smj", None),
    ("Lunda", "lun", None),
    ("Luo (Kenya and Tanzania)", "luo", None),
    ("Lushai", "lus", None),
    ("Luxembourgish", "ltz", "lb"),
    ("Macedo-Romanian", "rup", None),
    ("Macedonian", "mac/mkd", "mk"),
    ("Madurese", "mad", None),
    ("Magahi", "mag", None),
    ("Maithili", "mai", None),
    ("Makasar", "mak", None),
    ("Malagasy", "mlg", "mg"),
    ("Malay", "may/msa", "ms"),
    ("Malayalam", "mal", "ml"),
    ("Maldivian", "div", "dv"),
    ("Maltese", "mlt", "mt"),
    ("Manchu", "mnc", None),
    ("Mandar", "mdr", None),
    ("Mandingo", "man", None),
    ("Manipuri", "mni", None),
    ("Manobo languages", "mno", None),
    ("Manx", "glv", "gv"),
    ("Maori", "mao/mri", "mi"),
    ("Marathi", "mar", "mr"),
    ("Mari", "chm", None),
    ("Marshallese", "mah", "mh"),
    ("Marwari", "mwr", None),
    ("Masai", "mas", None),
    ("Mayan languages", "myn", None),
    ("Mende", "men", None),
    ("Mi'kmaq", "mic", None),
    ("Micmac", "mic", None),
    ("Minangkabau", "min", None),
    ("Mirandese", "mwl", None),
    ("Mohawk", "moh", None),
    ("Moksha", "mdf", None),
    ("Moldavian", "mol", "mo"),
    ("Mon-Khmer (Other)", "mkh", None),
    ("Mongo", "lol", None),
    ("Mongolian", "mon", "mn"),
    ("Mossi", "mos", None),
    ("Multiple languages", "mul", None),
    ("Munda languages", "mun", None),
    ("N'Ko", "nqo", None),
    ("Nahuatl", "nah", None),
    ("Nauru", "nau", "na"),
    ("Navaho", "nav", "nv"),
    ("Navajo", "nav", "nv"),
    ("Ndebele, North", "nde", "nd"),
    ("Ndebele, South", "nbl", "nr"),
    ("Ndonga", "ndo", "ng"),
    ("Neapolitan", "nap", None),
    ("Nepal Bhasa", "new", None),
    ("Nepali", "nep", "ne"),
    ("Newari", "new", None),
    ("Nias", "nia", None),
    ("Niger-Kordofanian (Other)", "nic", None),
    ("Nilo-Saharan (Other)", "ssa", None),
    ("Niuean", "niu", None),
    ("No linguistic content", "zxx", None),
    ("Nogai", "nog", None),
    ("Norse, Old", "non", None),
    ("North American Indian", "nai", None),
    ("North Ndebele", "nde", "nd"),
    ("Northern Frisian", "frr", None),
    ("Northern Sami", "sme", "se"),
    ("Northern Sotho", "nso", None),
    ("Norwegian", "nor", "no"),
    ("Norwegian Bokmål", "nob", "nb"),
    ("Norwegian Nynorsk", "nno", "nn"),
    ("Nubian languages", "nub", None),
    ("Nyamwezi", "nym", None),
    ("Nyanja", "nya", "ny"),
    ("Nyankole", "nyn", None),
    ("Nynorsk, Norwegian", "nno", "nn"),
    ("Nyoro", "nyo", None),
    ("Nzima", "nzi", None),
    ("Occitan (post 1500)", "oci", "oc"),
    ("Oirat", "xal", None),
    ("Ojibwa", "oji", "oj"),
    ("Old Bulgarian", "chu", "cu"),
    ("Old Church Slavonic", "chu", "cu"),
    ("Old Newari", "nwc", None),
    ("Old Slavonic", "chu", "cu"),
    ("Oriya", "ori", "or"),
    ("Oromo", "orm", "om"),
    ("Osage", "osa", None),
    ("Ossetian", "oss", "os"),
    ("Ossetic", "oss", "os"),
    ("Otomian languages", "oto", None),
    ("Pahlavi", "pal", None),
    ("Palauan", "pau", None),
    ("Pali", "pli", "pi"),
    ("Pampanga", "pam", None),
    ("Pangasinan", "pag", None),
    ("Panjabi", "pan", "pa"),
    ("Papiamento", "pap", None),
    ("Papuan (Other)", "paa", None),
    ("Pedi", "nso", None),
    ("Persian", "per/fas", "fa"),
    ("Persian, Old (ca.600-400 B.C.)", "peo", None),
    ("Philippine (Other)", "phi", None),
    ("Phoenician", "phn", None),
    ("Pilipino", "fil", None),
    ("Pohnpeian", "pon", None),
    ("Polish", "pol", "pl"),
    ("Portuguese", "por", "pt"),
    ("Prakrit languages", "pra", None),
    ("Provençal", "oci", "oc"),
    ("Provençal, Old (to 1500)", "pro", None),
    ("Punjabi", "pan", "pa"),
    ("Pushto", "pus", "ps"),
    ("Quechua", "que", "qu"),
    ("Raeto-Romance", "roh", "rm"),
    ("Rajasthani", "raj", None),
    ("Rapanui", "rap", None),
    ("Rarotongan", "rar", None),
    ("Reserved for local use", "qaa/qtz", None),
    ("Romance (Other)", "roa", None),
    ("Romanian", "rum/ron", "ro"),
    ("Romany", "rom", None),
    ("Rundi", "run", "rn"),
    ("Russian", "rus", "ru"),
    ("Salishan languages", "sal", None),
    ("Samaritan Aramaic", "sam", None),
    ("Sami languages (Other)", "smi", None),
    ("Samoan", "smo", "sm"),
    ("Sandawe", "sad", None),
    ("Sango", "sag", "sg"),
    ("Sanskrit", "san", "sa"),
    ("Santali", "sat", None),
    ("Sardinian", "srd", "sc"),
    ("Sasak", "sas", None),
    ("Saxon, Low", "nds", None),
    ("Scots", "sco", None),
    ("Scottish Gaelic", "gla", "gd"),
    ("Selkup", "sel", None),
    ("Semitic (Other)", "sem", None),
    ("Sepedi", "nso", None),
    ("Serbian", "scc/srp", "sr"),
    ("Serer", "srr", None),
    ("Shan", "shn", None),
    ("Shona", "sna", "sn"),
    ("Sichuan Yi", "iii", "ii"),
    ("Sicilian", "scn", None),
    ("Sidamo", "sid", None),
    ("Sign Languages", "sgn", None),
    ("Siksika", "bla", None),
    ("Sindhi", "snd", "sd"),
    ("Sinhala", "sin", "si"),
    ("Sinhalese", "sin", "si"),
    ("Sino-Tibetan (Other)", "sit", None),
    ("Siouan languages", "sio", None),
    ("Skolt Sami", "sms", None),
    ("Slave (Athapascan)", "den", None),
    ("Slavic (Other)", "sla", None),
    ("Slovak", "slo/slk", "sk"),
    ("Slovenian", "slv", "sl"),
    ("Sogdian", "sog", None),
    ("Somali", "som", "so"),
    ("Songhai", "son", None),
    ("Soninke", "snk", None),
    ("Sorbian languages", "wen", None),
    ("Sotho, Northern", "nso", None),
    ("Sotho, Southern", "sot", "st"),
    ("South American Indian (Other)", "sai", None),
    ("South Ndebele", "nbl", "nr"),
    ("Southern Altai", "alt", None),
    ("Southern Sami", "sma", None),
    ("Spanish", "spa", "es"),
    ("Sranan Togo", "srn", None),
    ("Sukuma", "suk", None),
    ("Sumerian", "sux", None),
    ("Sundanese", "sun", "su"),
    ("Susu", "sus", None),
    ("Swahili", "swa", "sw"),
    ("Swati", "ssw", "ss"),
    ("Swedish", "swe", "sv"),
    ("Swiss German", "gsw", None),
    ("Syriac", "syr", None),
    ("Tagalog", "tgl", "tl"),
    ("Tahitian", "tah", "ty"),
    ("Tai (Other)", "tai", None),
    ("Tajik", "tgk", "tg"),
    ("Tamashek", "tmh", None),
    ("Tamil", "tam", "ta"),
    ("Tatar", "tat", "tt"),
    ("Telugu", "tel", "te"),
    ("Tereno", "ter", None),
    ("Tetum", "tet", None),
    ("Thai", "tha", "th"),
    ("Tibetan", "tib/bod", "bo"),
    ("Tigre", "tig", None),
    ("Tigrinya", "tir", "ti"),
    ("Timne", "tem", None),
    ("Tiv", "tiv", None),
    ("tlhIngan-Hol", "tlh", None),
    ("Tlingit", "tli", None),
    ("Tok Pisin", "tpi", None),
    ("Tokelau", "tkl", None),
    ("Tonga (Nyasa)", "tog", None),
    ("Tonga (Tonga Islands)", "ton", "to"),
    ("Tsimshian", "tsi", None),
    ("Tsonga", "tso", "ts"),
    ("Tswana", "tsn", "tn"),
    ("Tumbuka", "tum", None),
    ("Tupi languages", "tup", None),
    ("Turkish", "tur", "tr"),
    ("Turkish, Ottoman (1500-1928)", "ota", None),
    ("Turkmen", "tuk", "tk"),
    ("Tuvalu", "tvl", None),
    ("Tuvinian", "tyv", None),
    ("Twi", "twi", "tw"),
    ("Udmurt", "udm", None),
    ("Ugaritic", "uga", None),
    ("Uighur", "uig", "ug"),
    ("Ukrainian", "ukr", "uk"),
    ("Umbundu", "umb", None),
    ("Uncoded languages", "mis", None),
    ("Undetermined", "und", None),
    ("Upper Sorbian", "hsb", None),
    ("Urdu", "urd", "ur"),
    ("Uyghur", "uig", "ug"),
    ("Uzbek", "uzb", "uz"),
    ("Vai", "vai", None),
    ("Valencian", "cat", "ca"),
    ("Venda", "ven", "ve"),
    ("Vietnamese", "vie", "vi"),
    ("Volapük", "vol", "vo"),
    ("Votic", "vot", None),
    ("Wakashan languages", "wak", None),
    ("Walamo", "wal", None),
    ("Walloon", "wln", "wa"),
    ("Waray", "war", None),
    ("Washo", "was", None),
    ("Welsh", "wel/cym", "cy"),
    ("Western Frisian", "fry", "fy"),
    ("Wolof", "wol", "wo"),
    ("Xhosa", "xho", "xh"),
    ("Yakut", "sah", None),
    ("Yao", "yao", None),
    ("Yapese", "yap", None),
    ("Yiddish", "yid", "yi"),
    ("Yoruba", "yor", "yo"),
    ("Yupik languages", "ypk", None),
    ("Zande", "znd", None),
    ("Zapotec", "zap", None),
    ("Zaza", "zza", None),
    ("Zazaki", "zza", None),
    ("Zenaga", "zen", None),
    ("Zhuang", "zha", "za"),
    ("Zulu", "zul", "zu"),
    ("Zuni", "zun", None),
)

# Bibliographic ISO-639-2 form (eg. "fre" => "French")
ISO639_2 = {}
for line in _ISO639:
    for key in line[1].split("/"):
        ISO639_2[key] = line[0]
del _ISO639



"""
Gimp image parser (XCF file, ".xcf" extension).

You can find informations about XCF file in Gimp source code. URL to read
CVS online:
  http://cvs.gnome.org/viewcvs/gimp/app/xcf/
  files xcf-read.c and xcf-load.c

Author: Victor Stinner
"""

from hachoir.parser import Parser
from hachoir.field import (StaticFieldSet, FieldSet, ParserError,
                           UInt8, UInt32, Enum, Float32, String, PascalString32, RawBytes)
from hachoir.parser.image.common import RGBA
from hachoir.core.endian import NETWORK_ENDIAN


class XcfCompression(FieldSet):
    static_size = 8
    COMPRESSION_NAME = {
        0: "None",
        1: "RLE",
        2: "Zlib",
        3: "Fractal"
    }

    def createFields(self):
        yield Enum(UInt8(self, "compression", "Compression method"), self.COMPRESSION_NAME)


class XcfResolution(StaticFieldSet):
    format = (
        (Float32, "xres", "X resolution in DPI"),
        (Float32, "yres", "Y resolution in DPI")
    )


class XcfTattoo(StaticFieldSet):
    format = ((UInt32, "tattoo", "Tattoo"),)


class LayerOffsets(StaticFieldSet):
    format = (
        (UInt32, "ofst_x", "Offset X"),
        (UInt32, "ofst_y", "Offset Y")
    )


class LayerMode(FieldSet):
    static_size = 32
    MODE_NAME = {
        0: "Normal",
        1: "Dissolve",
        2: "Behind",
        3: "Multiply",
        4: "Screen",
        5: "Overlay",
        6: "Difference",
        7: "Addition",
        8: "Subtract",
        9: "Darken only",
        10: "Lighten only",
        11: "Hue",
        12: "Saturation",
        13: "Color",
        14: "Value",
        15: "Divide",
        16: "Dodge",
        17: "Burn",
        18: "Hard light",
        19: "Soft light",
        20: "Grain extract",
        21: "Grain merge",
        22: "Color erase"
    }

    def createFields(self):
        yield Enum(UInt32(self, "mode", "Layer mode"), self.MODE_NAME)


class GimpBoolean(UInt32):

    def __init__(self, parent, name):
        UInt32.__init__(self, parent, name)

    def createValue(self):
        return 1 == UInt32.createValue(self)


class XcfUnit(StaticFieldSet):
    format = ((UInt32, "unit", "Unit"),)


class XcfParasiteEntry(FieldSet):

    def createFields(self):
        yield PascalString32(self, "name", "Name", strip="\0", charset="UTF-8")
        yield UInt32(self, "flags", "Flags")
        yield PascalString32(self, "data", "Data", strip=" \0", charset="UTF-8")


class XcfLevel(FieldSet):

    def createFields(self):
        yield UInt32(self, "width", "Width in pixel")
        yield UInt32(self, "height", "Height in pixel")
        yield UInt32(self, "offset", "Offset")
        offset = self["offset"].value
        if offset == 0:
            return
        data_offsets = []
        while (self.absolute_address + self.current_size) // 8 < offset:
            chunk = UInt32(self, "data_offset[]", "Data offset")
            yield chunk
            if chunk.value == 0:
                break
            data_offsets.append(chunk)
        if (self.absolute_address + self.current_size) // 8 != offset:
            raise ParserError("Problem with level offset.")
        previous = offset
        for chunk in data_offsets:
            data_offset = chunk.value
            size = data_offset - previous
            yield RawBytes(self, "data[]", size, "Data content of %s" % chunk.name)
            previous = data_offset


class XcfHierarchy(FieldSet):

    def createFields(self):
        yield UInt32(self, "width", "Width")
        yield UInt32(self, "height", "Height")
        yield UInt32(self, "bpp", "Bits/pixel")

        offsets = []
        while True:
            chunk = UInt32(self, "offset[]", "Level offset")
            yield chunk
            if chunk.value == 0:
                break
            offsets.append(chunk.value)
        for offset in offsets:
            padding = self.seekByte(offset, relative=False)
            if padding is not None:
                yield padding
            yield XcfLevel(self, "level[]", "Level")
#        yield XcfChannel(self, "channel[]", "Channel"))


class XcfChannel(FieldSet):

    def createFields(self):
        yield UInt32(self, "width", "Channel width")
        yield UInt32(self, "height", "Channel height")
        yield PascalString32(self, "name", "Channel name", strip="\0", charset="UTF-8")
        yield from readProperties(self)
        yield UInt32(self, "hierarchy_ofs", "Hierarchy offset")
        yield XcfHierarchy(self, "hierarchy", "Hierarchy")

    def createDescription(self):
        return 'Channel "%s"' % self["name"].value


class XcfLayer(FieldSet):

    def createFields(self):
        yield UInt32(self, "width", "Layer width in pixels")
        yield UInt32(self, "height", "Layer height in pixels")
        yield Enum(UInt32(self, "type", "Layer type"), XcfFile.IMAGE_TYPE_NAME)
        yield PascalString32(self, "name", "Layer name", strip="\0", charset="UTF-8")
        for prop in readProperties(self):
            yield prop

        # --
        # TODO: Hack for Gimp 1.2 files
        # --

        yield UInt32(self, "hierarchy_ofs", "Hierarchy offset")
        yield UInt32(self, "mask_ofs", "Layer mask offset")
        padding = self.seekByte(self["hierarchy_ofs"].value, relative=False)
        if padding is not None:
            yield padding
        yield XcfHierarchy(self, "hierarchy", "Hierarchy")
        # TODO: Read layer mask if needed: self["mask_ofs"].value != 0

    def createDescription(self):
        return 'Layer "%s"' % self["name"].value


class XcfParasites(FieldSet):

    def createFields(self):
        size = self["../size"].value * 8
        while self.current_size < size:
            yield XcfParasiteEntry(self, "parasite[]", "Parasite")


class XcfProperty(FieldSet):
    PROP_COMPRESSION = 17
    PROP_RESOLUTION = 19
    PROP_PARASITES = 21
    TYPE_NAME = {
        0: "End",
        1: "Colormap",
        2: "Active layer",
        3: "Active channel",
        4: "Selection",
        5: "Floating selection",
        6: "Opacity",
        7: "Mode",
        8: "Visible",
        9: "Linked",
        10: "Lock alpha",
        11: "Apply mask",
        12: "Edit mask",
        13: "Show mask",
        14: "Show masked",
        15: "Offsets",
        16: "Color",
        17: "Compression",
        18: "Guides",
        19: "Resolution",
        20: "Tattoo",
        21: "Parasites",
        22: "Unit",
        23: "Paths",
        24: "User unit",
        25: "Vectors",
        26: "Text layer flags",
    }

    handler = {
        6: RGBA,
        7: LayerMode,
        8: GimpBoolean,
        9: GimpBoolean,
        10: GimpBoolean,
        11: GimpBoolean,
        12: GimpBoolean,
        13: GimpBoolean,
        15: LayerOffsets,
        17: XcfCompression,
        19: XcfResolution,
        20: XcfTattoo,
        21: XcfParasites,
        22: XcfUnit
    }

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        self._size = (8 + self["size"].value) * 8

    def createFields(self):
        yield Enum(UInt32(self, "type", "Property type"), self.TYPE_NAME)
        yield UInt32(self, "size", "Property size")

        size = self["size"].value
        if 0 < size:
            cls = self.handler.get(self["type"].value, None)
            if cls:
                yield cls(self, "data", size=size * 8)
            else:
                yield RawBytes(self, "data", size, "Data")

    def createDescription(self):
        return "Property: %s" % self["type"].display


def readProperties(parser):
    while True:
        prop = XcfProperty(parser, "property[]")
        yield prop
        if prop["type"].value == 0:
            return


class XcfFile(Parser):
    PARSER_TAGS = {
        "id": "xcf",
        "category": "image",
        "file_ext": ("xcf",),
        "mime": ("image/x-xcf", "application/x-gimp-image"),
        # header+empty property+layer offset+channel offset
        "min_size": (26 + 8 + 4 + 4) * 8,
        "magic": (
            (b'gimp xcf file\0', 0),
            (b'gimp xcf v002\0', 0),
        ),
        "description": "Gimp (XCF) picture"
    }
    endian = NETWORK_ENDIAN
    IMAGE_TYPE_NAME = {
        0: "RGB",
        1: "Gray",
        2: "Indexed"
    }

    def validate(self):
        if self.stream.readBytes(0, 14) not in (b'gimp xcf file\0', b'gimp xcf v002\0'):
            return "Wrong signature"
        return True

    def createFields(self):
        # Read signature
        yield String(self, "signature", 14, "Gimp picture signature (ends with nul byte)", charset="ASCII")

        # Read image general informations (width, height, type)
        yield UInt32(self, "width", "Image width")
        yield UInt32(self, "height", "Image height")
        yield Enum(UInt32(self, "type", "Image type"), self.IMAGE_TYPE_NAME)
        for prop in readProperties(self):
            yield prop

        # Read layer offsets
        layer_offsets = []
        while True:
            chunk = UInt32(self, "layer_offset[]", "Layer offset")
            yield chunk
            if chunk.value == 0:
                break
            layer_offsets.append(chunk.value)

        # Read channel offsets
        channel_offsets = []
        while True:
            chunk = UInt32(self, "channel_offset[]", "Channel offset")
            yield chunk
            if chunk.value == 0:
                break
            channel_offsets.append(chunk.value)

        # Read layers
        for index, offset in enumerate(layer_offsets):
            if index + 1 < len(layer_offsets):
                size = (layer_offsets[index + 1] - offset) * 8
            else:
                size = None
            padding = self.seekByte(offset, relative=False)
            if padding:
                yield padding
            yield XcfLayer(self, "layer[]", size=size)

        # Read channels
        for index, offset in enumerate(channel_offsets):
            if index + 1 < len(channel_offsets):
                size = (channel_offsets[index + 1] - offset) * 8
            else:
                size = None
            padding = self.seekByte(offset, relative=False)
            if padding is not None:
                yield padding
            yield XcfChannel(self, "channel[]", "Channel", size=size)



"""
7zip file parser

Informations:
- File 7zformat.txt of 7-zip SDK:
  http://www.7-zip.org/sdk.html

Author: Olivier SCHWAB
Creation date: 6 december 2006

Updated by: Robert Xiao
Date: February 26 2011
"""

from hachoir.parser import Parser
from hachoir.field import (Field, FieldSet, ParserError, CString,
                           Enum, Bit, Bits, UInt8, UInt32, UInt64,
                           Bytes, RawBytes, TimestampWin64)
from hachoir.stream import StringInputStream
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal
from hachoir.core.tools import alignValue, humanFilesize, makePrintable
from hachoir.parser.common.msdos import MSDOSFileAttr32

try:
    from pylzma import decompress as lzmadecompress
    has_lzma = True
except ImportError:
    has_lzma = False


class SZUInt64(Field):
    """
    Variable length UInt64, where the first byte gives both the number of bytes
    needed and the upper byte value.
    """

    def __init__(self, parent, name, max_size=None, description=None):
        Field.__init__(self, parent, name, size=8, description=description)
        value = 0
        addr = self.absolute_address
        mask = 0x80
        firstByte = parent.stream.readBits(addr, 8, LITTLE_ENDIAN)
        for i in range(8):
            addr += 8
            if not (firstByte & mask):
                value += ((firstByte & (mask - 1)) << (8 * i))
                break
            value |= (parent.stream.readBits(
                addr, 8, LITTLE_ENDIAN) << (8 * i))
            mask >>= 1
            self._size += 8
        self.createValue = lambda: value


kEnd = 0x00
kHeader = 0x01
kArchiveProperties = 0x02
kAdditionalStreamsInfo = 0x03
kMainStreamsInfo = 0x04
kFilesInfo = 0x05
kPackInfo = 0x06
kUnPackInfo = 0x07
kSubStreamsInfo = 0x08
kSize = 0x09
kCRC = 0x0A
kFolder = 0x0B
kCodersUnPackSize = 0x0C
kNumUnPackStream = 0x0D
kEmptyStream = 0x0E
kEmptyFile = 0x0F
kAnti = 0x10
kName = 0x11
kCreationTime = 0x12
kLastAccessTime = 0x13
kLastWriteTime = 0x14
kWinAttributes = 0x15
kComment = 0x16
kEncodedHeader = 0x17
kStartPos = 0x18
kDummy = 0x19

PROP_IDS = {
    kEnd: 'kEnd',
    kHeader: 'kHeader',
    kArchiveProperties: 'kArchiveProperties',
    kAdditionalStreamsInfo: 'kAdditionalStreamsInfo',
    kMainStreamsInfo: 'kMainStreamsInfo',
    kFilesInfo: 'kFilesInfo',
    kPackInfo: 'kPackInfo',
    kUnPackInfo: 'kUnPackInfo',
    kSubStreamsInfo: 'kSubStreamsInfo',
    kSize: 'kSize',
    kCRC: 'kCRC',
    kFolder: 'kFolder',
    kCodersUnPackSize: 'kCodersUnPackSize',
    kNumUnPackStream: 'kNumUnPackStream',
    kEmptyStream: 'kEmptyStream',
    kEmptyFile: 'kEmptyFile',
    kAnti: 'kAnti',
    kName: 'kName',
    kCreationTime: 'kCreationTime',
    kLastAccessTime: 'kLastAccessTime',
    kLastWriteTime: 'kLastWriteTime',
    kWinAttributes: 'kWinAttributes',
    kComment: 'kComment',
    kEncodedHeader: 'kEncodedHeader',
    kStartPos: 'kStartPos',
    kDummy: 'kDummy',
}

PROP_DESC = {
    kEnd: 'End-of-header marker',

    kHeader: 'Archive header',

    kArchiveProperties: 'Archive properties',

    kAdditionalStreamsInfo: 'AdditionalStreamsInfo',
    kMainStreamsInfo: 'MainStreamsInfo',
    kFilesInfo: 'FilesInfo',

    kPackInfo: 'PackInfo',
    kUnPackInfo: 'UnPackInfo',
    kSubStreamsInfo: 'SubStreamsInfo',

    kSize: 'Size',
    kCRC: 'CRC',

    kFolder: 'Folder',

    kCodersUnPackSize: 'CodersUnPackSize',
    kNumUnPackStream: 'NumUnPackStream',

    kEmptyStream: 'EmptyStream',
    kEmptyFile: 'EmptyFile',
    kAnti: 'Anti',

    kName: 'Name',
    kCreationTime: 'CreationTime',
    kLastAccessTime: 'LastAccessTime',
    kLastWriteTime: 'LastWriteTime',
    kWinAttributes: 'WinAttributes',
    kComment: 'Comment',

    kEncodedHeader: 'Encoded archive header',

    kStartPos: 'Unknown',
    kDummy: 'Dummy entry',
}


def ReadNextByte(self):
    return self.stream.readBits(self.absolute_address + self.current_size, 8, self.endian)


def PropID(self, name):
    return Enum(UInt8(self, name), PROP_IDS)


class SevenZipBitVector(FieldSet):

    def __init__(self, parent, name, num, has_all_byte=False, **args):
        FieldSet.__init__(self, parent, name, **args)
        self.has_all_byte = has_all_byte
        self.num = num

    def createFields(self):
        if self.has_all_byte:
            yield Enum(UInt8(self, "all_defined"), {0: 'False', 1: 'True'})
            if self['all_defined'].value:
                return
        nbytes = alignValue(self.num, 8) // 8
        ctr = 0
        for i in range(nbytes):
            for j in reversed(range(8)):
                yield Bit(self, "bit[%d]" % (ctr + j))
            ctr += 8

    def isAllDefined(self):
        return self.has_all_byte and self['all_defined'].value

    def isDefined(self, index):
        if self.isAllDefined():
            return True
        return self['bit[%d]' % index].value

    def createValue(self):
        if self.isAllDefined():
            return list(range(self.num))
        return [i for i in range(self.num) if self['bit[%d]' % i].value]

    def createDisplay(self):
        if self.isAllDefined():
            return 'all'
        return ','.join(str(i) for i in self.value)


class ArchiveProperty(FieldSet):

    def createFields(self):
        yield PropID(self, "id")
        size = SZUInt64(self, "size")
        yield size
        if size.value:
            yield RawBytes(self, "data", size.value)

    def createDescription(self):
        return self['id'].display


class ArchiveProperties(FieldSet):

    def createFields(self):
        yield PropID(self, "id")
        while not self.eof:
            uid = ReadNextByte(self)
            if uid == kEnd:
                yield PropID(self, "end_marker")
                break
            yield ArchiveProperty(self, "prop[]")


class Digests(FieldSet):

    def __init__(self, parent, name, num_digests, digest_desc=None, desc=None):
        FieldSet.__init__(self, parent, name, desc)
        self.num_digests = num_digests
        if digest_desc is None:
            self.digest_desc = ['stream %d' % i for i in range(num_digests)]
        else:
            self.digest_desc = digest_desc

    def createFields(self):
        yield PropID(self, "id")
        definearr = SevenZipBitVector(
            self, "defined", self.num_digests, has_all_byte=True)
        yield definearr
        for index in definearr.value:
            yield textHandler(UInt32(self, "digest[]",
                                     "Digest for %s" % self.digest_desc[index]), hexadecimal)


class PackInfo(FieldSet):

    def createFields(self):
        yield PropID(self, "id")

        yield SZUInt64(self, "pack_pos", "File offset to the packed data")
        num = SZUInt64(self, "num_pack_streams", "Number of packed streams")
        yield num

        while not self.eof:
            uid = ReadNextByte(self)
            if uid == kEnd:
                yield PropID(self, "end_marker")
                break
            elif uid == kSize:
                yield PropID(self, "size_marker")
                for index in range(num.value):
                    yield SZUInt64(self, "pack_size[]")
            elif uid == kCRC:
                yield Digests(self, "digests", num.value)
            else:
                raise ParserError("Unexpected ID (%i)" % uid)


METHODS = {
    b"\0": "Copy",
    b"\3": "Delta",
    b"\4": "x86_BCJ",
    b"\5": "PowerPC",
    b"\6": "IA64",
    b"\7": "ARM_LE",
    b"\x08": "ARMT_LE",  # thumb
    b"\x09": "SPARC",
    b"\x21": "LZMA2",
    b"\2\3\2": "Common-Swap-2",
    b"\2\3\4": "Common-Swap-4",
    b"\3\1\1": "7z-LZMA",
    b"\3\3\1\3": "7z-Branch-x86-BCJ",
    b"\3\3\1\x1b": "7z-Branch-x86-BCJ2",
    b"\3\3\2\5": "7z-Branch-PowerPC-BE",
    b"\3\3\3\1": "7z-Branch-Alpha-LE",
    b"\3\3\4\1": "7z-Branch-IA64-LE",
    b"\3\3\5\1": "7z-Branch-ARM-LE",
    b"\3\3\6\5": "7z-Branch-M68-BE",
    b"\3\3\7\1": "7z-Branch-ARMT-LE",
    b"\3\3\x08\5": "7z-Branch-SPARC-BE",
    b"\3\4\1": "7z-PPMD",
    b"\3\x7f\1": "7z-Experimental",
    b"\4\0": "Reserved",
    b"\4\1\0": "Zip-Copy",
    b"\4\1\1": "Zip-Shrink",
    b"\4\1\6": "Zip-Implode",
    b"\4\1\x08": "Zip-Deflate",
    b"\4\1\x09": "Zip-Deflate64",
    b"\4\1\x10": "Zip-BZip2",
    b"\4\1\x14": "Zip-LZMA",
    b"\4\1\x60": "Zip-JPEG",
    b"\4\1\x61": "Zip-WavPack",
    b"\4\1\x62": "Zip-PPMD",
    b"\4\1\x63": "Zip-wzAES",
    b"\4\2\2": "BZip2",
    b"\4\3\1": "RAR-15",
    b"\4\3\2": "RAR-20",
    b"\4\3\3": "RAR-29",
    b"\4\4\1": "Arj3",
    b"\4\4\2": "Arj4",
    b"\4\5": "Z",
    b"\4\6": "LZH",
    b"\4\7": "7z-Reserved",
    b"\4\x08": "CAB",
    b"\4\x09\1": "NSIS-Deflate",
    b"\4\x09\2": "NSIS-BZip2",
    b"\6\0": "Crypto-Reserved",
    b"\6\1\x00": "Crypto-AES128-ECB",
    b"\6\1\x01": "Crypto-AES128-CBC",
    b"\6\1\x02": "Crypto-AES128-CFB",
    b"\6\1\x03": "Crypto-AES128-OFB",
    b"\6\1\x04": "Crypto-AES128-CTR",
    b"\6\1\x40": "Crypto-AES192-ECB",
    b"\6\1\x41": "Crypto-AES192-CBC",
    b"\6\1\x42": "Crypto-AES192-CFB",
    b"\6\1\x43": "Crypto-AES192-OFB",
    b"\6\1\x44": "Crypto-AES192-CTR",
    b"\6\1\x80": "Crypto-AES256-ECB",
    b"\6\1\x81": "Crypto-AES256-CBC",
    b"\6\1\x82": "Crypto-AES256-CFB",
    b"\6\1\x83": "Crypto-AES256-OFB",
    b"\6\1\x84": "Crypto-AES256-CTR",
    b"\6\1\xc0": "Crypto-AES-ECB",
    b"\6\1\xc1": "Crypto-AES-CBC",
    b"\6\1\xc2": "Crypto-AES-CFB",
    b"\6\1\xc3": "Crypto-AES-OFB",
    b"\6\1\xc4": "Crypto-AES-CTR",
    b"\6\7": "Crypto-Reserved",
    b"\6\x0f": "Crypto-Reserved",
    b"\6\xf0": "Crypto-Misc",
    b"\6\xf1\1\1": "Crypto-Zip",
    b"\6\xf1\3\2": "Crypto-RAR-Unknown",
    b"\6\xf1\3\3": "Crypto-RAR-29",  # AES128
    b"\6\xf1\7\1": "Crypto-7z",  # AES256
    b"\7\0": "Hash-None",
    b"\7\1": "Hash-CRC",
    b"\7\2": "Hash-SHA1",
    b"\7\3": "Hash-SHA256",
    b"\7\4": "Hash-SHA384",
    b"\7\5": "Hash-SHA512",
    b"\7\xf0": "Hash-Misc",
    b"\7\xf1\3\3": "Hash-RAR-29",  # modified SHA1
    b"\7\xf1\7\1": "Hash-7z",  # SHA256
}


class Coder(FieldSet):

    def createFields(self):
        yield Bits(self, "id_size", 4)
        yield Bit(self, "is_not_simple", "If unset, stream setup is simple")
        yield Bit(self, "has_attribs", "Are there compression properties attached?")
        yield Bit(self, "unused[]")
        yield Bit(self, "is_not_last_method", "Are there more methods after this one in the alternative method list?")
        size = self['id_size'].value
        if size > 0:
            yield Enum(RawBytes(self, "id", size), METHODS)
        if self['is_not_simple'].value:
            yield SZUInt64(self, "num_stream_in")
            yield SZUInt64(self, "num_stream_out")
            self.info("Streams: IN=%u    OUT=%u" %
                      (self["num_stream_in"].value, self["num_stream_out"].value))
        if self['has_attribs'].value:
            size = SZUInt64(self, "properties_size")
            yield size
            yield RawBytes(self, "properties", size.value)

    def _get_num_streams(self, direction):
        if self['is_not_simple'].value:
            return self['num_stream_%s' % direction].value
        return 1
    in_streams = property(lambda self: self._get_num_streams('in'))
    out_streams = property(lambda self: self._get_num_streams('out'))


class CoderList(FieldSet):

    def createFields(self):
        while not self.eof:
            field = Coder(self, "coder[]")
            yield field
            if not field['is_not_last_method'].value:
                break


class BindPairInfo(FieldSet):

    def createFields(self):
        # 64 bits values then cast to 32 in fact
        yield SZUInt64(self, "in_index")
        yield SZUInt64(self, "out_index")
        self.info("Indexes: IN=%u   OUT=%u" %
                  (self["in_index"].value, self["out_index"].value))


class Folder(FieldSet):

    def createFields(self):
        yield SZUInt64(self, "num_coders")
        num = self["num_coders"].value
        self.info("Folder: %u codecs" % num)

        in_streams = out_streams = 0

        # Coder info
        for index in range(num):
            ci = CoderList(self, "coders[]")
            yield ci
            in_streams += ci['coder[0]'].in_streams
            out_streams += ci['coder[0]'].out_streams
        self._in_streams = in_streams
        self._out_streams = out_streams

        # Bind pairs
        self.info("out streams: %u" % out_streams)
        for index in range(out_streams - 1):
            yield BindPairInfo(self, "bind_pair[]")

        # Packed streams
        # @todo: Actually find mapping
        packed_streams = in_streams - out_streams + 1
        if packed_streams > 1:
            for index in range(packed_streams):
                yield SZUInt64(self, "pack_stream[]")

    def _get_num_streams(self, direction):
        list(self)
        return getattr(self, '_' + direction + '_streams')
    in_streams = property(lambda self: self._get_num_streams('in'))
    out_streams = property(lambda self: self._get_num_streams('out'))


class UnpackInfo(FieldSet):

    def createFields(self):
        yield PropID(self, "id")

        yield PropID(self, "folder_marker")
        assert self['folder_marker'].value == kFolder
        yield SZUInt64(self, "num_folders")

        # Get generic info
        num = self["num_folders"].value
        self.info("%u folders" % num)
        yield UInt8(self, "is_external")

        if self['is_external'].value:
            yield SZUInt64(self, "folder_data_offset", "Offset to folder data within data stream")
        else:
            # Read folder items
            for folder_index in range(num):
                yield Folder(self, "folder[]")

        yield PropID(self, "unpacksize_marker")
        assert self['unpacksize_marker'].value == kCodersUnPackSize
        for folder_index in range(num):
            folder = self["folder[%u]" % folder_index]
            for index in range(folder.out_streams):
                yield SZUInt64(self, "unpack_size[%d][%d]" % (folder_index, index))

        # Extract digests
        while not self.eof:
            uid = ReadNextByte(self)
            if uid == kEnd:
                yield PropID(self, "end_marker")
                break
            elif uid == kCRC:
                yield Digests(self, "digests", num)
            else:
                raise ParserError("Unexpected ID (%i)" % uid)


class SubStreamInfo(FieldSet):

    def createFields(self):
        yield PropID(self, "id")
        num_folders = self['../unpack_info/num_folders'].value
        num_unpackstreams = [1] * num_folders
        while not self.eof:
            uid = ReadNextByte(self)
            if uid == kEnd:
                yield PropID(self, "end_marker")
                break
            elif uid == kNumUnPackStream:
                yield PropID(self, "num_unpackstream_marker")
                for i in range(num_folders):
                    field = SZUInt64(self, "num_unpackstreams[]")
                    yield field
                    num_unpackstreams[i] = field.value
            elif uid == kSize:
                yield PropID(self, "size_marker")
                for i in range(num_folders):
                    # The last substream's size is the stream size minus the
                    # other substreams.
                    for j in range(num_unpackstreams[i] - 1):
                        yield SZUInt64(self, "unpack_size[%d][%d]" % (i, j))
            elif uid == kCRC:
                digests = []
                for i in range(num_folders):
                    if num_unpackstreams[i] == 1 and 'digests' in self['../unpack_info']:
                        continue
                    for j in range(num_unpackstreams[i]):
                        digests.append('folder %i, stream %i' % (i, j))
                yield Digests(self, "digests", len(digests), digests)
            else:
                raise ParserError("Unexpected ID (%i)" % uid)


class StreamsInfo(FieldSet):

    def createFields(self):
        yield PropID(self, "id")
        while not self.eof:
            uid = ReadNextByte(self)
            if uid == kEnd:
                yield PropID(self, "end")
                break
            elif uid == kPackInfo:
                yield PackInfo(self, "pack_info", PROP_DESC[uid])
            elif uid == kUnPackInfo:
                yield UnpackInfo(self, "unpack_info", PROP_DESC[uid])
            elif uid == kSubStreamsInfo:
                yield SubStreamInfo(self, "substreams_info", PROP_DESC[uid])
            else:
                raise ParserError("Unexpected ID (%i)" % uid)


class EncodedHeader(StreamsInfo):
    pass


class EmptyStreamProperty(FieldSet):

    def createFields(self):
        yield PropID(self, "id")
        yield SZUInt64(self, "size")
        yield SevenZipBitVector(self, "vec", self['../num_files'].value)

    def createValue(self):
        return self['vec'].value

    def createDisplay(self):
        return self['vec'].display


class EmptyFileProperty(FieldSet):

    def createFields(self):
        yield PropID(self, "id")
        yield SZUInt64(self, "size")
        empty_streams = self['../empty_streams/vec'].value
        yield SevenZipBitVector(self, "vec", len(empty_streams))

    def createValue(self):
        empty_streams = self['../empty_streams/vec'].value
        return [empty_streams[i] for i in self['vec'].value]

    def createDisplay(self):
        return ','.join(str(i) for i in self.value)


class FileTimeProperty(FieldSet):

    def createFields(self):
        yield PropID(self, "id")
        yield SZUInt64(self, "size")
        definearr = SevenZipBitVector(
            self, "defined", self['../num_files'].value, has_all_byte=True)
        yield definearr
        yield UInt8(self, "is_external")
        if self['is_external'].value:
            yield SZUInt64(self, "folder_data_offset", "Offset to folder data within data stream")
        else:
            for index in definearr.value:
                yield TimestampWin64(self, "timestamp[%d]" % index)


class FileNames(FieldSet):

    def createFields(self):
        yield PropID(self, "id")
        yield SZUInt64(self, "size")
        yield UInt8(self, "is_external")
        if self['is_external'].value:
            yield SZUInt64(self, "folder_data_offset", "Offset to folder data within data stream")
        else:
            for index in range(self['../num_files'].value):
                yield CString(self, "name[%d]" % index, charset="UTF-16-LE")


class FileAttributes(FieldSet):

    def createFields(self):
        yield PropID(self, "id")
        yield SZUInt64(self, "size")
        definearr = SevenZipBitVector(
            self, "defined", self['../num_files'].value, has_all_byte=True)
        yield definearr
        yield UInt8(self, "is_external")
        if self['is_external'].value:
            yield SZUInt64(self, "folder_data_offset", "Offset to folder data within data stream")
        else:
            for index in definearr.value:
                yield MSDOSFileAttr32(self, "attributes[%d]" % index)


class FilesInfo(FieldSet):

    def createFields(self):
        yield PropID(self, "id")
        yield SZUInt64(self, "num_files")
        while not self.eof:
            uid = ReadNextByte(self)
            if uid == kEnd:
                yield PropID(self, "end_marker")
                break
            elif uid == kEmptyStream:
                yield EmptyStreamProperty(self, "empty_streams")
            elif uid == kEmptyFile:
                yield EmptyFileProperty(self, "empty_files")
            elif uid == kAnti:
                yield EmptyFileProperty(self, "anti_files")
            elif uid == kCreationTime:
                yield FileTimeProperty(self, "creation_time")
            elif uid == kLastAccessTime:
                yield FileTimeProperty(self, "access_time")
            elif uid == kLastWriteTime:
                yield FileTimeProperty(self, "modified_time")
            elif uid == kName:
                yield FileNames(self, "filenames")
            elif uid == kWinAttributes:
                yield FileAttributes(self, "attributes")
            elif uid == kDummy:
                yield ArchiveProperty(self, "dummy[]")
            else:
                yield ArchiveProperty(self, "prop[]")


class Header(FieldSet):

    def createFields(self):
        yield PropID(self, "id")
        while not self.eof:
            uid = ReadNextByte(self)
            if uid == kEnd:
                yield PropID(self, "end")
                break
            elif uid == kArchiveProperties:
                yield ArchiveProperties(self, "props", PROP_DESC[uid])
            elif uid == kAdditionalStreamsInfo:
                yield StreamsInfo(self, "additional_streams", PROP_DESC[uid])
            elif uid == kMainStreamsInfo:
                yield StreamsInfo(self, "main_streams", PROP_DESC[uid])
            elif uid == kFilesInfo:
                yield FilesInfo(self, "files_info", PROP_DESC[uid])
            else:
                raise ParserError("Unexpected ID %u" % uid)


class NextHeader(FieldSet):

    def __init__(self, parent, name, desc="Next header"):
        FieldSet.__init__(self, parent, name, desc)
        self._size = 8 * self["/signature/start_hdr/next_hdr_size"].value

    def createFields(self):
        uid = ReadNextByte(self)
        if uid == kHeader:
            yield Header(self, "header", PROP_DESC[uid])
        elif uid == kEncodedHeader:
            yield EncodedHeader(self, "encoded_hdr", PROP_DESC[uid])
        else:
            raise ParserError("Unexpected ID %u" % uid)


class NextHeaderParser(Parser):
    PARSER_TAGS = {
    }
    endian = LITTLE_ENDIAN

    def createFields(self):
        uid = ReadNextByte(self)
        if uid == kHeader:
            yield Header(self, "header", PROP_DESC[uid])
        elif uid == kEncodedHeader:
            yield EncodedHeader(self, "encoded_hdr", PROP_DESC[uid])
        else:
            raise ParserError("Unexpected ID %u" % uid)

    def validate(self):
        return True


class CompressedData(Bytes):

    def __init__(self, parent, name, length, decompressor, description=None,
                 parser=None, filename=None, mime_type=None, parser_class=None):
        if filename:
            if not isinstance(filename, str):
                filename = makePrintable(filename, "ISO-8859-1")
            if not description:
                description = 'File "%s" (%s)' % (
                    filename, humanFilesize(length))
        Bytes.__init__(self, parent, name, length, description)
        self.setupInputStream(decompressor, parser,
                              filename, mime_type, parser_class)

    def setupInputStream(self, decompressor, parser, filename, mime_type, parser_class):
        def createInputStream(cis, **args):
            tags = args.setdefault("tags", [])
            if parser_class:
                tags.append(("class", parser_class))
            if parser is not None:
                tags.append(("id", parser.PARSER_TAGS["id"]))
            if mime_type:
                tags.append(("mime", mime_type))
            if filename:
                tags.append(("filename", filename))
            return StringInputStream(decompressor(self.value), **args)
        self.setSubIStream(createInputStream)


def get_header_decompressor(self):
    unpack_info = self['/next_hdr/encoded_hdr/unpack_info']
    assert unpack_info['num_folders'].value == 1
    coder = unpack_info['folder[0]/coders[0]/coder[0]']
    method = METHODS[coder['id'].value]
    if method == 'Copy':
        return lambda data: data
    elif method == '7z-LZMA' and has_lzma:
        props = coder['properties'].value
        length = unpack_info['unpack_size[0][0]'].value
        return lambda data: lzmadecompress(props + data, maxlength=length)


def get_header_field(self, name, size, description=None):
    decompressor = get_header_decompressor(self)
    if decompressor is None:
        return RawBytes(self, name, size, description=description)
    return CompressedData(self, name, size, decompressor, description=description, parser_class=NextHeaderParser)


class Body(FieldSet):

    def __init__(self, parent, name, desc="Body data"):
        FieldSet.__init__(self, parent, name, desc)
        self._size = 8 * self["/signature/start_hdr/next_hdr_offset"].value

    def createFields(self):
        if "encoded_hdr" in self["/next_hdr"]:
            pack_size = sum([s.value for s in self.array(
                "/next_hdr/encoded_hdr/pack_info/pack_size")])
            body_size = self["/next_hdr/encoded_hdr/pack_info/pack_pos"].value
            if body_size:
                yield RawBytes(self, "compressed_data", body_size, "Compressed data")
            # Here we could check if copy method was used to "compress" it,
            # but this never happens, so just output "compressed file info"
            yield get_header_field(self, "compressed_file_info", pack_size,
                                   "Compressed file information")
            size = (self._size // 8) - pack_size - body_size
            if size > 0:
                yield RawBytes(self, "unknown_data", size)
        elif "header" in self["/next_hdr"]:
            yield RawBytes(self, "compressed_data", self._size // 8, "Compressed data")


class StartHeader(FieldSet):
    static_size = 160

    def createFields(self):
        yield textHandler(UInt64(self, "next_hdr_offset",
                                 "Next header offset"), hexadecimal)
        yield UInt64(self, "next_hdr_size", "Next header size")
        yield textHandler(UInt32(self, "next_hdr_crc",
                                 "Next header CRC"), hexadecimal)


class SignatureHeader(FieldSet):
    static_size = 96 + StartHeader.static_size

    def createFields(self):
        yield Bytes(self, "signature", 6, "Signature Header")
        yield UInt8(self, "major_ver", "Archive major version")
        yield UInt8(self, "minor_ver", "Archive minor version")
        yield textHandler(UInt32(self, "start_hdr_crc",
                                 "Start header CRC"), hexadecimal)
        yield StartHeader(self, "start_hdr", "Start header")


class SevenZipParser(Parser):
    MAGIC = b"7z\xbc\xaf\x27\x1c"
    PARSER_TAGS = {
        "id": "7zip",
        "category": "archive",
        "file_ext": ("7z",),
        "mime": ("application/x-7z-compressed",),
        "min_size": 32 * 8,
        "magic": ((MAGIC, 0),),
        "description": "Compressed archive in 7z format"
    }
    endian = LITTLE_ENDIAN

    def createFields(self):
        yield SignatureHeader(self, "signature", "Signature Header")
        yield Body(self, "body_data")
        yield NextHeader(self, "next_hdr")

    def validate(self):
        if self.stream.readBytes(0, len(self.MAGIC)) != self.MAGIC:
            return "Invalid signature"
        return True

    def createContentSize(self):
        size = self["/signature/start_hdr/next_hdr_offset"].value * 8
        size += self["/signature/start_hdr/next_hdr_size"].value * 8
        size += SignatureHeader.static_size
        return size
